"""Functions to interact with PhagesDB"""

import json
import pathlib
from urllib import request, error

from pdm_utils.classes import genome
from pdm_utils.constants import constants


def parse_phage_name(data_dict):
    """Retrieve Phage Name from PhagesDB.

    :param data_dict: Dictionary of data retrieved from PhagesDB.
    :type data_dict: dict
    :returns: Name of the phage.
    :rtype: str
    """
    try:
        phage_name = data_dict["phage_name"]
    except KeyError:
        phage_name = ""
    return phage_name


def parse_cluster(data_dict):
    """Retrieve Cluster from PhagesDB.

    If the phage is clustered, 'pcluster' is a dictionary, and one key is
    the Cluster data (Cluster or 'Singleton').
    If for some reason no Cluster info is added at the time
    the genome is added to PhagesDB, 'pcluster' may automatically be
    set to NULL, which gets converted to "Unclustered" during retrieval.
    In the MySQL database NULL means Singleton, and the long form
    "Unclustered" is invalid due to its character length,
    so this value is converted to 'UNK' ('Unknown').

    :param data_dict: Dictionary of data retrieved from PhagesDB.
    :type data_dict: dict
    :returns: Cluster of the phage.
    :rtype: str
    """
    try:
        if data_dict["pcluster"] is None:
            cluster = "UNK"
        else:
            cluster = data_dict["pcluster"]["cluster"]
    except KeyError:
        cluster = ""
    return cluster


def parse_subcluster(data_dict):
    """Retrieve Subcluster from PhagesDB.

    If for some reason no cluster info is added at the time
    the genome is added to PhagesDB, 'psubcluster' may automatically be
    set to NULL, which gets returned as None.
    If the phage is a Singleton, 'psubcluster' is None.
    If the phage is clustered but not subclustered, 'psubcluster' is None.
    If the phage is clustered and subclustered, 'psubcluster'
    is a dictionary, and one key is the Subcluster data.

    :param data_dict: Dictionary of data retrieved from PhagesDB.
    :type data_dict: dict
    :returns: Subcluster of the phage.
    :rtype: str
    """
    try:
        if data_dict["psubcluster"] is None:
            subcluster = "none"
        else:
            subcluster = data_dict["psubcluster"]["subcluster"]
    except KeyError:
        subcluster = ""
    return subcluster


def parse_host_genus(data_dict):
    """Retrieve host_genus from PhagesDB.

    :param data_dict: Dictionary of data retrieved from PhagesDB.
    :type data_dict: dict
    :returns: Host genus of the phage.
    :rtype: str
    """
    try:
        host_genus = data_dict["isolation_host"]["genus"]
    except KeyError:
        host_genus = ""
    return host_genus


def parse_accession(data_dict):
    """Retrieve Accession from PhagesDB.

    :param data_dict: Dictionary of data retrieved from PhagesDB.
    :type data_dict: dict
    :returns: Accession of the phage.
    :rtype: str
    """
    try:
        accession = data_dict["genbank_accession"]
    except KeyError:
        accession = ""
    return accession


def parse_fasta_filename(data_dict):
    """Retrieve fasta filename from PhagesDB.

    :param data_dict: Dictionary of data retrieved from PhagesDB.
    :type data_dict: dict
    :returns: Name of the fasta file for the phage.
    :rtype: str
    """
    try:
        fastafile_url = data_dict["fasta_file"]
    except KeyError:
        fastafile_url = ""
    return fastafile_url


def retrieve_url_data(url):
    """Retrieve fasta file from PhagesDB.

    :param url: URL for data to be retrieved.
    :type url: str
    :returns: Data from the URL.
    :rtype: str
    """
    try:
        # gcontext = ssl.SSLContext(ssl.PROTOCOL_TLSv1) ==> required for
        # urllib2.URLError: <urlopen error [SSL: CERTIFICATE_VERIFY_FAILED]
        # certificate verify failed (_ssl.c:590)> ==> creating new TLS
        # context tells urllib2 to ignore certificate chain
        # NOTE this is BAD SECURITY, prone to man-in-the-middle attacks
        rqst = request.Request(url)
        with request.urlopen(rqst) as response:
            data = response.read()
            data = data.decode("utf-8")
    except error.HTTPError:
        print(f"Unable to retrieve data from {url}")
        data = ""
    return data


def parse_fasta_data(fasta_data):
    """Parses data returned from a fasta-formatted file.

    :param fasta_data: Data from a fasta file.
    :type fasta_data: str
    :returns:
        tuple (header, sequence)
        WHERE
        header(str) is the first line parsed from the parsed file.
        sequence(str) is the nucleotide sequence parsed from the file.
    :rtype: tuple
    """
    # All sequence rows in the fasta file may not have equal widths,
    # so some processing of the data is required. If you split by newline,
    # the header is retained in the first list element.
    split_fasta_data = fasta_data.split('\n')
    header = ""
    sequence = ""
    if len(split_fasta_data) > 1:
        header = split_fasta_data[0]
        if header[0] == ">":
            header = header[1:]  # Remove '>' symbol.
        header = header.strip()  # Remove any whitespace
        for index in range(1, len(split_fasta_data)):
            # Strip off potential whitespace before appending, such as '\r'.
            sequence = sequence + split_fasta_data[index].strip()
    return header, sequence


def parse_genome_data(data_dict, gnm_type="", seq=False):
    """Parses a dictionary of PhagesDB genome data into a pdm_utils Genome object.

    :param data_dict: Dictionary of data retrieved from PhagesDB.
    :type data_dict: dict
    :param gnm_type: Identifier for the type of genome.
    :type gnm_type: str
    :param seq: Indicates whether the genome sequence should be retrieved.
    :type seq: bool
    :returns: A pdm_utils Genome object with the parsed data.
    :rtype: Genome
    """
    gnm = genome.Genome()
    gnm.type = gnm_type

    # Phage Name, PhageID
    phage_name = parse_phage_name(data_dict)
    gnm.name = phage_name
    gnm.set_id(value=phage_name)

    # Host
    host_genus = parse_host_genus(data_dict)
    gnm.set_host_genus(host_genus, "empty_string")

    # Accession
    accession = parse_accession(data_dict)
    gnm.set_accession(accession, "empty_string")

    # Cluster
    cluster = parse_cluster(data_dict)
    gnm.set_cluster(cluster)

    # Subcluster
    subcluster = parse_subcluster(data_dict)
    gnm.set_subcluster(subcluster)

    # Fasta file URL
    fastafile_url = parse_fasta_filename(data_dict)
    fastafile_url_path = pathlib.Path(fastafile_url)
    gnm.set_filename(fastafile_url_path)

    # Fasta file record
    # if fastafile_url != "":
    if fastafile_url != "" and seq is True:
        fasta_file = retrieve_url_data(fastafile_url)

        # TODO unit test - not sure how to test this, since this function
        # retrieves and parses files from PhagesDB.
        # Genome sequence and parsed record
        if fasta_file != "":
            header, seq = parse_fasta_data(fasta_file)
            gnm.set_sequence(seq)
            gnm.description = header
            gnm.parse_description()

    gnm.misc = data_dict
    return gnm


def retrieve_genome_data(phage_url):
    """Retrieve all data from PhagesDB for a specific phage.

    :param phage_url: URL for data pertaining to a specific phage.
    :type phage_url: str
    :returns: Dictionary of data parsed from the URL.
    :rtype: dict
    """
    try:
        data_json = request.urlopen(phage_url)
        data_dict = json.loads(data_json.read())
        # TODO should data_dict.close() be called after retrieving data?
    except error.HTTPError:
        data_dict = {}
    return data_dict


def construct_phage_url(phage_name):
    """Create URL to retrieve phage-specific data from PhagesDB.

    :param phage_name: Name of the phage of interest.
    :type phage_name: str
    :returns: URL pertaining to the phage.
    :rtype: str
    """
    phage_url = constants.API_PREFIX + phage_name + constants.API_SUFFIX
    return phage_url


def get_genome(phage_id, gnm_type="", seq=False):
    """Get genome data from PhagesDB.

    :param phage_id: The name of the phage to be retrieved from PhagesDB.
    :type phage_id: str
    :param gnm_type: Identifier for the type of genome.
    :type gnm_type: str
    :param seq: Indicates whether the genome sequence should be retrieved.
    :type seq: bool
    :returns:
        A pdm_utils Genome object with the parsed data.
        If not genome is retrieved, None is returned.
    :rtype: Genome
    """
    phage_url = construct_phage_url(phage_id)
    data_dict = retrieve_genome_data(phage_url)
    if len(data_dict.keys()) != 0:
        gnm = parse_genome_data(data_dict, gnm_type=gnm_type, seq=seq)
    else:
        gnm = None
    return gnm


# TODO unittest.
def get_phagesdb_data(url):
    """Retrieve all sequenced genome data from PhagesDB.

    :param url: URL to connect to PhagesDB API.
    :type url: str
    :returns:
        List of dictionaries, where each dictionary contains
        data for each phage. If a problem is encountered during retrieval,
        an empty list is returned.
    :rtype: list
    """
    data_json = request.urlopen(url)
    # Response is a bytes object that json.loads can't read without first
    # being decoded to a UTF-8 string.
    data_dict = json.loads(data_json.read().decode("utf-8"))
    data_json.close()

    # Returned dict:
    # Keys:
    # count = # of phages
    # results = list of sequenced phage data
    data_list = data_dict["results"]
    diff = len(data_list) - data_dict["count"]
    if diff != 0:
        print("\nUnable to retrieve all phage data from PhagesDB due to "
              "default parameters.")
        data_list = []
    return data_list


# TODO unittest.
def parse_genomes_dict(data_dict, gnm_type="", seq=False):
    """Returns a dictionary of pdm_utils Genome objects

    :param data_dict:
        Dictionary of dictionaries.
        Key = PhageID.
        Value = Dictionary of genome data retrieved from PhagesDB.
    :type data_dict: dict
    :param gnm_type: Identifier for the type of genome.
    :type gnm_type: str
    :param seq: Indicates whether the genome sequence should be retrieved.
    :type seq: bool
    :returns:
        Dictionary of pdm_utils Genome object.
        Key = PhageID.
        Value = Genome object.
    :rtype: dict
    """
    genome_dict = {}
    for key in data_dict.keys():
        gnm = parse_genome_data(data_dict[key], gnm_type=gnm_type, seq=seq)
        genome_dict[gnm.id] = gnm
    return genome_dict


def retrieve_data_list(url):
    """Retrieve list of data from PhagesDB.

    :param url: A URL from which to retrieve data.
    :type url: str
    :returns: A list of data retrieved from the URL.
    :rtype: list
    """
    try:
        data_json = request.urlopen(url)
        data_list = json.loads(data_json.read())
    except error.HTTPError:
        data_list = []
    return data_list


def create_host_genus_set(url=constants.API_HOST_GENERA):
    """Create a set of host genera currently in PhagesDB.

    :param url: A URL from which to retrieve host genus data.
    :type url: str
    :returns: All unique host genera listed on PhagesDB.
    :rtype: set
    """
    try:
        output = retrieve_data_list(url)
    except error.HTTPError:
        output = []
    host_genera_set = set()
    for genus_dict in output:
        try:
            host_genera_set.add(genus_dict["genus_name"])
        except KeyError:
            pass
    return host_genera_set


def create_cluster_subcluster_sets(url=constants.API_CLUSTERS):
    """Create sets of clusters and subclusters currently in PhagesDB.

    :param url: A URL from which to retrieve cluster and subcluster data.
    :type url: str
    :returns:
        tuple (cluster_set, subcluster_set)
        WHERE
        cluster_set(set) is a set of all unique clusters on PhagesDB.
        subcluster_set(set) is a set of all unique subclusters on PhagesDB.
    :rtype: tuple
    """
    try:
        output = retrieve_data_list(url)
    except error.HTTPError:
        output = []
    cluster_set = set()
    subcluster_set = set()
    for data in output:
        try:
            cluster_set.add(data["cluster"])  # This set contains 'Singleton'.
            try:
                subclusters_list = data["subclusters_set"]
                subcluster_set = subcluster_set | set(subclusters_list)
            except KeyError:
                pass
        except KeyError:
            pass
    return cluster_set, subcluster_set



"""Represents a structure to contain directions for how to parse and import
genomes into a MySQL database."""

from pdm_utils.classes import evaluation
from pdm_utils.functions import basic




class ImportTicket:

    # Initialize all attributes:
    def __init__(self):

        # Initialize all non-calculated attributes:
        self.id = None # Unique identifier
        self.type = "" # Add, Replace
        self.phage_id = ""
        self.eval_mode = ""
        self.description_field = ""
        self.eval_flags = {} # Dictionary of evaluation flags.

        # Attributes used to populate Genome objects for
        # 'update', 'add', and 'replace' ticket types.
        self.data_retrieve = set() # Data that should be retrieved from PhagesDB.
        self.data_retain = set() # Data that should be retained from the MySQL database.
        self.data_add = set() # Data to be added to genome from ticket.
        self.data_parse = set() # Data that should be parsed from the flat file.
        self.data_dict = {} # All data from import table.

        # Used to check the structure of the ticket data.
        self.evaluations = []



    # When parsing data from an import ticket, some fields should be
    # lowercase regardless of the content, while other fields should
    # be lowercase only if they are set to 'none' or 'retrieve'.
    def set_type(self, value):
        """Set the ticket type.

        :param value: Value to be set as the type.
        :type value: str
        """
        self.type = value.lower()

    def set_description_field(self, value):
        """Set the description_field.

        :param value: Value to be set as the description_field.
        :type value: str
        """
        self.description_field = value.lower()

    def set_eval_mode(self, value):
        """Set the eval_mode.

        :param value: Value to be set as the eval_mode.
        :type value: str
        """
        self.eval_mode = value.lower()


    # Evaluations
    def set_eval(self, eval_id, definition, result, status):
        """Constructs and adds an Evaluation object to the evaluations list.

        :param eval_id: Unique identifier for the evaluation.
        :type eval_id: str
        :param definition: Description of the evaluation.
        :type definition: str
        :param result: Description of the outcome of the evaluation.
        :type result: str
        :param status: Outcome of the evaluation.
        :type status: str
        """
        evl = evaluation.Evaluation(eval_id, definition, result, status)
        self.evaluations.append(evl)


    def check_attribute(self, attribute, check_set, expect=False, eval_id=None,
                        success="correct", fail="error", eval_def=None):
        """Check that the id is valid.

        :param attribute: Name of the ImportTicket object attribute to evaluate.
        :type attribute: str
        :param check_set: Set of reference ids.
        :type check_set: set
        :param expect:
            Indicates whether the attribute value is expected to be present
            in the check set.
        :type expect: bool
        :param eval_id: Unique identifier for the evaluation.
        :type eval_id: str
        :param success: Default status if the outcome is a success.
        :type success: str
        :param fail: Default status if the outcome is not a success.
        :type fail: str
        :param eval_def: Description of the evaluation.
        :type eval_def: str
        """
        try:
            test = True
            value1 = getattr(self, attribute)
        except:
            test = False
            value1 = None
        if test:
            value2 = basic.check_value_expected_in_set(
                        value1, check_set, expect)
            if value2:
                result = f"The {attribute} is valid."
                status = success
            else:
                result = f"The {attribute} is not valid."
                status = fail
        else:
            result = f"The {attribute} was not evaluated."
            status = "untested"
        definition = f"Check the {attribute} attribute."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_eval_flags(self, expect=True, eval_id=None,
                         success="correct", fail="error", eval_def=None):
        """Check that the eval_flags is valid.

        :param expect:
            Indicates whether the eval_flags is expected to contain data.
        :type expect: bool
        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        keys = len(self.eval_flags)
        msg = f"There are {keys} evaluation flags present, which is "
        if keys > 0 and expect:
            output = True
        elif keys == 0 and not expect:
            output = True
        else:
            output = False

        if output:
            result = msg + "expected."
            status = success
        else:
            result = msg + "not expected."
            status = fail
        definition = "Check if there are evaluation flags."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_compatible_type_and_data_retain(self, eval_id=None,
                          success="correct", fail="error", eval_def=None):
        """Check if the ticket type and data_retain are compatible.

        If the ticket type is 'add', then the data_retain set is not
        expected to have any data.

        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        msg = (f"The ticket type is '{self.type}' and "
               f"there are {len(self.data_retain)} values in the database "
               "that are set to be retained, which is ")
        if (self.type == "add" and len(self.data_retain) > 0):
            result = msg + "not expected."
            status = fail
        else:
            result = msg + "expected."
            status = success
        definition = ("Check if the ticket type and data_retain setting "
                      "are compatible.")
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)


    def check_valid_data_source(self, ref_set_attr, check_set, eval_id=None,
                                success="correct", fail="error", eval_def=None):
        """Check that the values in the specified attribute are valid.

        :param ref_set_attr:
            Name of the data_dict in the ticket to be evaluated
            (data_add, data_retain, data_retrieve, data_parse)
        :type ref_set_attr: str
        :param check_set: Set of valid field names.
        :type check_set: set
        :param eval_id: same as for check_attribute().
        :param success: same as for check_attribute().
        :param fail: same as for check_attribute().
        :param eval_def: same as for check_attribute().
        """
        if ref_set_attr == "data_add":
            ref_set = self.data_add
        elif ref_set_attr == "data_retain":
            ref_set = self.data_retain
        elif ref_set_attr == "data_retrieve":
            ref_set = self.data_retrieve
        elif ref_set_attr == "data_parse":
            ref_set = self.data_parse
        else:
            ref_set = None
        if ref_set is not None:
            invalid_values = ref_set - check_set
            msg = f"The '{ref_set_attr}' field is ."
            if len(invalid_values) == 0:
                result = "populated correctly."
                status = success
            else:
                invalid_string = basic.join_strings(invalid_values,
                                                    delimiter=", ")
                result = (
                    "not populated correctly. The following "
                    f"values are not permitted in '{ref_set_attr}': "
                    f"{invalid_string}")
                status = fail
        else:
            result = f"'{ref_set_attr}' is not a valid attribute to be evaluated."
            status = fail

        definition = f"Check if {ref_set_attr} field is correctly populated."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)



"""Represents a structure to contain results of an evaluation.
"""

class Evaluation:
    def __init__(self, id="", definition="", result="", status=""):
        self.id = id # Unique identifier for the specific evaluation.
        self.definition = definition # Description of what was evaluated.
        self.status = status # untested, ignored, correct, warning, error.
        self.result = result # Customized message reporting details of
                             # the evaluation.

    def __str__(self):
        return (f"Evaluation ID: {self.id}. "
                f"Status: {self.status}. "
                f"Definition: {self.definition} "
                f"Result: {self.result}")



"""Functions to retrieve phage genome annotation data."""

import re

from sqlalchemy import select

from pdm_utils.functions import basic
from pdm_utils.functions import querying


# ANNOTATION RETRIEVAL
# -----------------------------------------------------------------------------
def get_genes_from_pham(alchemist, pham):
    gene_obj = alchemist.metadata.tables["gene"]

    geneid_obj = gene_obj.c.GeneID
    phamid_obj = gene_obj.c.PhamID

    genes_query = select([geneid_obj]).where(phamid_obj == pham).distinct()

    genes = querying.first_column(alchemist.engine, genes_query)
    return genes


def get_distinct_phams_from_genes(alchemist, geneids):
    gene_obj = alchemist.metadata.tables["gene"]

    geneid_obj = gene_obj.c.GeneID
    phamid_obj = gene_obj.c.PhamID

    phams_query = select([phamid_obj]).distinct()

    phams = querying.first_column(alchemist.engine, phams_query,
                                  in_column=geneid_obj, values=geneids)
    return phams


def get_distinct_annotations_from_genes(alchemist, geneids):
    gene_obj = alchemist.metadata.tables["gene"]

    geneid_obj = gene_obj.c.GeneID
    notes_obj = gene_obj.c.Notes

    notes_query = select([notes_obj]).distinct()

    bytes_annotations = querying.first_column(alchemist.engine, notes_query,
                                              in_column=geneid_obj,
                                              values=geneids)

    annotations = basic.convert_to_decoded(bytes_annotations)
    return annotations


def get_phams_from_genes(alchemist, geneids):
    gene_obj = alchemist.metadata.tables["gene"]

    geneid_obj = gene_obj.c.GeneID
    phamid_obj = gene_obj.c.PhamID

    pham_query = select([phamid_obj])

    phams = []
    for gene in geneids:
        pham = alchemist.engine.execute(pham_query.where(geneid_obj == gene))\
                                                                    .scalar()
        phams.append(pham)

    return phams


def get_annotations_from_genes(alchemist, geneids):
    gene_obj = alchemist.metadata.tables["gene"]

    geneid_obj = gene_obj.c.GeneID
    notes_obj = gene_obj.c.Notes

    notes_query = select([notes_obj])

    annotations = []
    for gene in geneids:
        note = alchemist.engine.execute(notes_query.where(geneid_obj == gene))\
                                                                    .scalar()

        if note is not None:
            note = note.decode("utf-8")
        annotations.append(note)

    return annotations


def get_count_phams_in_genes(alchemist, geneids, incounts=None):
    phams = get_phams_from_genes(alchemist, geneids)

    pham_histogram = {}
    if incounts is not None:
        pham_histogram = incounts

    basic.increment_histogram(phams, pham_histogram)
    return pham_histogram


def get_count_annotations_in_genes(alchemist, geneids, incounts=None):
    annotations = get_annotations_from_genes(alchemist, geneids)

    annotation_counts = {}
    if incounts is not None:
        annotation_counts = incounts

    basic.increment_histogram(annotations, annotation_counts)
    return annotation_counts


def get_count_annotations_in_pham(alchemist, pham, incounts=None):
    genes = get_genes_from_pham(alchemist, pham)
    count_annotations = get_count_annotations_in_genes(alchemist, genes,
                                                       incounts=incounts)
    return count_annotations


def get_relative_gene(alchemist, geneid, pos):
    gene_obj = alchemist.metadata.tables["gene"]

    geneid_obj = gene_obj.c.GeneID

    geneid_format = re.compile("[\w\W]+_CDS_[0-9]+")
    if not re.match(geneid_format, geneid) is None:
        parsed_geneid = re.split("_", geneid)
    else:
        raise ValueError(f"Passed GeneID {geneid} "
                         "is not of the proper GeneID format")

    gene_num = int(parsed_geneid[2])
    rel_gene_pos = gene_num + pos

    rel_geneid = "_".join(parsed_geneid[:2] + [str(rel_gene_pos)])
    geneid_query = select([geneid_obj]).where(geneid_obj == rel_geneid)
    rel_geneid = alchemist.engine.execute(geneid_query).scalar()

    return rel_geneid


def get_adjacent_genes(alchemist, gene):
    left = get_relative_gene(alchemist, gene, -1)
    right = get_relative_gene(alchemist, gene, 1)

    return (left, right)


def get_genes_adjacent_to_pham(alchemist, pham):
    genes = get_genes_from_pham(alchemist, pham)

    left_genes = []
    right_genes = []
    for gene in genes:
        adjacent_genes = get_adjacent_genes(alchemist, gene)

        if not adjacent_genes[0] is None:
            left_genes.append(adjacent_genes[0])
        if not adjacent_genes[1] is None:
            right_genes.append(adjacent_genes[1])

    return (left_genes, right_genes)


def get_distinct_adjacent_phams(alchemist, pham):
    adjacent_genes = get_genes_adjacent_to_pham(alchemist, pham)

    left_phams = get_phams_from_genes(alchemist, adjacent_genes[0])
    right_phams = get_phams_from_genes(alchemist, adjacent_genes[1])

    return (left_phams, right_phams)


def get_count_adjacent_phams_to_pham(alchemist, pham, incounts=None):
    adjacent_genes = get_genes_adjacent_to_pham(alchemist, pham)

    adjacent_phams = ({}, {})
    if incounts is not None:
        adjacent_phams = incounts

    left_in = adjacent_phams[0]
    get_count_phams_in_genes(alchemist, adjacent_genes[0], incounts=left_in)

    right_in = adjacent_phams[1]
    get_count_phams_in_genes(alchemist, adjacent_genes[1], incounts=right_in)

    return adjacent_phams


def get_count_adjacent_annotations_to_pham(alchemist, pham, incounts=None):
    adjacent_genes = get_genes_adjacent_to_pham(alchemist, pham)

    adjacent_annotations = ({}, {})
    if incounts is not None:
        adjacent_annotations = incounts

    left_in = adjacent_annotations[0]
    get_count_annotations_in_genes(alchemist, adjacent_genes[0],
                                   incounts=left_in)

    right_in = adjacent_annotations[1]
    get_count_annotations_in_genes(alchemist, adjacent_genes[1],
                                   incounts=right_in)

    return adjacent_annotations



"""
Pipeline to push files to a server using SFTP.
"""

import argparse
import getpass
import pathlib
import sys

import paramiko

from pdm_utils.functions import basic
from pdm_utils.functions import configfile


# TODO unittest.
def parse_args(unparsed_args):
    """
    Verify the correct arguments are selected for uploading to the server.
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("-d", "--directory",
                        type=pathlib.Path,
                        help="path to a folder whose contents will be "
                             "uploaded to the server")
    parser.add_argument("-f", "--file",
                        type=pathlib.Path,
                        help="path to a file to be uploaded to the server")
    parser.add_argument("-l", "--log_file",
                        type=pathlib.Path,
                        default=pathlib.Path("/tmp/paramiko.log"),
                        help="path to a file where paramiko output should "
                             "be logged")
    parser.add_argument("-s", "--server_host",
                        type=str,
                        default=None,
                        help="server hostname")
    parser.add_argument("-k", "--key-file",
                        type=pathlib.Path,
                        help="path to a key-file to use for server login")
    parser.add_argument("-rd", "--remote_directory",
                        type=pathlib.Path,
                        default=None,
                        help="path to server directory where uploaded file(s) "
                             "should go")
    parser.add_argument("-c", "--config_file",
                        type=pathlib.Path,
                        default=None,
                        help="path to local file containing server login data")

    args = parser.parse_args(unparsed_args)
    return args


# TODO test.
def get_files(directory, file, ignore):
    """
    Get the list of file(s) that need to be uploaded.

    :param directory: (optional) directory containing files for upload
    :type: directory: pathlib.Path
    :param file: (optional) file to upload
    :type file: pathlib.Path
    :param ignore: file(s) to ignore during upload process
    :type ignore: set
    :return: file_list
    """
    file_list = []

    if directory is not None:
        directory = basic.set_path(directory, kind="dir", expect=True)
        folder_files = basic.identify_contents(directory, kind="file",
                                               ignore_set=ignore)
        file_list.extend(folder_files)

    if file is not None:
        file = basic.set_path(file, kind="file", expect=True)
        file_list.append(file)

    return file_list


def upload(sftp_client, destination, files):
    """
    Try to upload the file(s).

    :param sftp_client: an open SFTPClient
    :type sftp_client: paramiko.SFTPClient
    :param destination: remote file directory to upload to
    :type destination: pathlib.Path
    :param files: the file(s) to upload
    :type files: list of pathlib.Path
    :return: successes, failures
    """
    successes, failures = list(), list()

    for local_file in files:
        name = local_file.name
        print(f"Uploading {name}...")
        remote_file = destination.joinpath(name)
        try:
            sftp_client.put(str(local_file), str(remote_file))
            successes.append(name)
        except OSError:
            failures.append(name)

    return successes, failures


# TODO unittest.
def main(unparsed_args):
    """
    Driver function for the push pipeline.

    :param unparsed_args: the command-line arguments given to this
                          pipeline's caller (likely pdm_utils.__main__)
    :type unparsed_args: list
    """
    # Parse the command line args with argparse
    # Assumed command line arg structure:
    # python3 -m pdm_utils.run <pipeline> <additional args...>
    # sys.argv:      [0]            [1]         [2...]
    args = parse_args(unparsed_args[2:])

    # Parse config file if one was given
    config = configfile.build_complete_config(args.config_file)
    server_host = config["upload_server"]["host"]
    remote_dir = config["upload_server"]["dest"]
    user = config["upload_server"]["user"]
    password = config["upload_server"]["password"]

    # Command line hostname and destination override config file defaults
    if args.server_host is not None:
        server_host = args.server_host
    if args.remote_directory is not None:
        remote_dir = args.remote_directory
    key_file = args.key_file

    # Can't upload files to unknown host or destination
    if server_host is None or remote_dir is None:
        print("No hostname and/or remote directory provided. Unable "
              "to upload file(s).")
        sys.exit(1)

    # Get the list of files to upload
    file_list = get_files(args.directory, args.file, ignore={".DS_Store"})

    if len(file_list) > 0:
        if args.log_file:
            paramiko.util.log_to_file(args.log_file)

        # Keep track of failed uploads
        failures = list()

        # Setup paramiko connection to server
        with paramiko.Transport(server_host) as transport:
            # Context manager so Transport will be closed automatically
            if user and key_file:
                # Username and key-file is preferred authentication method
                try:
                    key = paramiko.RSAKey.from_private_key_file(key_file)
                except paramiko.ssh_exception.SSHException:
                    print(f"'{key_file}' is not a valid RSA key file")
                    sys.exit(1)
                try:
                    transport.connect(username=user, pkey=key)
                except paramiko.ssh_exception.AuthenticationException:
                    print(f"Authentication failed with user '{user}' and "
                          f"key-file '{key_file}'")
                    sys.exit(1)
            elif user and password:
                # Username and password from config is next preferred method
                try:
                    transport.connect(username=user, password=password)
                except paramiko.ssh_exception.AuthenticationException:
                    p = f"{password[0]}{'*' * (len(password)-2)}{password[-1]}"
                    print(f"Authentication failed with user '{user}' and "
                          f"password '{p}'")
                    sys.exit(1)
            else:
                # Finally, get username and password from command line
                user = getpass.getpass(f"Enter username for {server_host}: ")
                password = getpass.getpass(f"Enter password for "
                                           f"{user}@{server_host}: ")
                try:
                    transport.connect(username=user, password=password)
                except paramiko.ssh_exception.AuthenticationException:
                    p = f"{password[0]}{'*' * (len(password)-2)}{password[-1]}"
                    print(f"Authentication failed with user '{user}' and "
                          f"password '{p}'")
                    sys.exit(1)
            with paramiko.SFTPClient.from_transport(transport) as sftp_client:
                # Context manager so SFTPClient will be closed automatically
                for local_file in file_list:
                    # sftp_client.put requires remote filename not remote dir
                    remote_file = remote_dir.joinpath(local_file.name)
                    try:
                        print(f"Uploading {str(local_file)}...")
                        sftp_client.put(str(local_file), str(remote_file))
                    except OSError:
                        failures.append(local_file)
        for file in failures:
            print(f"Could not upload {str(file)}")
    else:
        print("No files to upload")



"""Functions to interact with, use, and parse genomic data from
GenBank-formatted flat files."""

import pathlib
from collections import OrderedDict
from datetime import datetime

from Bio import SeqIO
from Bio.SeqFeature import (
                SeqFeature, CompoundLocation, FeatureLocation, ExactPosition)
from Bio import Alphabet
from Bio.Alphabet import IUPAC
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from pdm_utils.classes import genome, cds, trna, tmrna, source
from pdm_utils.constants import constants


def retrieve_genome_data(filepath):
    """Retrieve data from a GenBank-formatted flat file.

    :param filepath:
        Path to GenBank-formatted flat file that will be parsed
        using Biopython.
    :type filepath: Path
    :returns:
        If there is only one record, a Biopython SeqRecord of parsed data.
        If the file cannot be parsed, or if there are multiple records,
        None value is returned.
    :rtype: SeqRecord
    """
    try:
        seqrecords = list(SeqIO.parse(filepath, "genbank"))
    except:
        seqrecords = []
    # filename = filepath.split("/")[-1]
    if len(seqrecords) == 0:
        print(f"There are no records in {filepath.name}.")
        seqrecord = None
    elif len(seqrecords) > 1:
        print(f"There are multiple records in {filepath.name}.")
        seqrecord = None
    else:
        seqrecord = seqrecords[0]
        return seqrecord


def parse_coordinates(seqfeature):
    """Parse the boundary coordinates from a GenBank-formatted flat file.

    The functions takes a Biopython SeqFeature object containing data
    that was parsed from the feature in the flat file. Parsing these
    coordinates can be tricky.
    There can be more than one set of coordinates if it is
    a compound location. Only features with 1 or 2 open reading frames
    (parts) are correctly parsed. Also, the boundaries may not be precise;
    instead they may be open or fuzzy. Non-precise coordinates are
    converted to '-1'. If the strand is undefined, the coordinates
    are converted to '-1' and parts is set to '0'. If an incorrect
    data type is provided, coorindates are set to '-1' and parts is
    set to '0'.

    :param seqfeature: Biopython SeqFeature
    :type seqfeature: SeqFeature
    :returns:
        tuple (start, stop, parts)
        WHERE
        start(int) is the first coordinate, regardless of strand.
        stop(int) is the second coordinate, regardless of strand.
        parts(int) is the number of open reading frames that define
        the feature.
    """
    start_position = None
    stop_position = None
    start = -1
    stop = -1
    parts = 0

    if (isinstance(seqfeature.location, FeatureLocation) or
            isinstance(seqfeature.location, CompoundLocation)):

        if seqfeature.strand is None:
            pass
        elif isinstance(seqfeature.location, FeatureLocation):
            parts = 1
            start_position = seqfeature.location.start
            stop_position = seqfeature.location.end
        elif isinstance(seqfeature.location, CompoundLocation):
            parts = len(seqfeature.location.parts)

            # Skip this compound seqfeature if it is comprised of more
            # than two features (tricky to parse).
            if parts == 2:

                # Retrieve compound seqfeature positions based on strand.
                if seqfeature.strand == 1:
                    start_position = seqfeature.location.parts[0].start
                    stop_position = seqfeature.location.parts[1].end
                elif seqfeature.strand == -1:
                    start_position = seqfeature.location.parts[1].start
                    stop_position = seqfeature.location.parts[0].end
                else:
                    pass
            else:
                pass
        else:
            pass
    else:
        pass
    if isinstance(start_position, ExactPosition):
        start = int(start_position)
    if isinstance(stop_position, ExactPosition):
        stop = int(stop_position)
    return (start, stop, parts)


def parse_cds_seqfeature(seqfeature):
    """Parse data from a Biopython CDS SeqFeature object into a Cds object.

    :param seqfeature: Biopython SeqFeature
    :type seqfeature: SeqFeature
    :param genome_id:

        An identifier for the genome in which the seqfeature
        is defined.

    :type genome_id: str
    :returns: A  pdm_utils Cds object
    :rtype: Cds
    """
    cds_ftr = cds.Cds()
    cds_ftr.seqfeature = seqfeature

    try:
        locus_tag = seqfeature.qualifiers["locus_tag"][0]
    except:
        locus_tag = ""
    finally:
        cds_ftr.set_locus_tag(locus_tag, delimiter=None)

    cds_ftr.set_orientation(seqfeature.strand, "fr_short", case=True)
    cds_ftr.start, cds_ftr.stop, cds_ftr.parts = parse_coordinates(seqfeature)

    # Coordinate format for GenBank flat file features parsed by Biopython
    # are 0-based half open intervals.
    cds_ftr.coordinate_format = "0_half_open"

    # For translation, convert it to a Biopython Seq object.
    try:
        translation = seqfeature.qualifiers["translation"][0]
    except:
        translation = ""
    finally:
        translation = Seq(translation, Alphabet.IUPAC.protein)
        cds_ftr.set_translation(translation)

    cds_ftr.set_nucleotide_length(translation=True)

    try:
        translation_table = seqfeature.qualifiers["transl_table"][0]
    except:
        translation_table = 0
    finally:
        cds_ftr.set_translation_table(translation_table)

    try:
        product = seqfeature.qualifiers["product"][0]
    except:
        product = ""
    finally:
        cds_ftr.set_description_field("product", product)

    try:
        function = seqfeature.qualifiers["function"][0]
    except:
        function = ""
    finally:
        cds_ftr.set_description_field("function", function)

    try:
        note = seqfeature.qualifiers["note"][0]
    except:
        note = ""
    finally:
        cds_ftr.set_description_field("note", note)

    try:
        gene = seqfeature.qualifiers["gene"][0]
    except:
        gene = ""
    finally:
        cds_ftr.set_gene(gene)

    cds_ftr.set_name()
    return cds_ftr


# TODO: Christian unit test
def parse_trna_seqfeature(seqfeature):
    """
    Parse data from a Biopython tRNA SeqFeature object into a
    Trna object.
    :param seqfeature: Biopython SeqFeature
    :type seqfeature: SeqFeature
    :returns: a pdm_utils Trna object
    :rtype: Trna
    """
    trna_ftr = trna.Trna()
    trna_ftr.seqfeature = seqfeature

    try:
        locus_tag = seqfeature.qualifiers["locus_tag"][0]
    except (KeyError, IndexError):
        locus_tag = ""
    finally:
        trna_ftr.set_locus_tag(locus_tag, delimiter=None)

    trna_ftr.set_orientation(seqfeature.strand, "fr_short", True)
    trna_ftr.start, trna_ftr.stop, trna_ftr.parts = parse_coordinates(
                                                                    seqfeature)

    # Coordinate format for GenBank flat file features parsed by Biopython
    # are 0-based half open intervals.
    trna_ftr.coordinate_format = "0_half_open"

    trna_ftr.set_nucleotide_length(use_seq=True)

    try:
        product = seqfeature.qualifiers["product"][0]
    except (KeyError, IndexError):
        product = ""
    finally:
        trna_ftr.product = product

    try:
        note = seqfeature.qualifiers["note"][0]
    except (KeyError, IndexError):
        note = ""
    finally:
        trna_ftr.note = note

    try:
        gene = seqfeature.qualifiers["gene"][0]
    except (KeyError, IndexError):
        gene = ""
    finally:
        trna_ftr.gene = gene

    trna_ftr.set_name()
    return trna_ftr


# TODO: Christian unit test
def parse_tmrna_seqfeature(seqfeature):
    """
    Parses data from a BioPython tmRNA SeqFeature object into a
    Tmrna object.
    :param seqfeature: BioPython SeqFeature
    :type seqfeature: SeqFeature
    :return: pdm_utils Tmrna object
    :rtype: Tmrna
    """
    tmrna_ftr = tmrna.Tmrna()
    tmrna_ftr.seqfeature = seqfeature

    try:
        locus_tag = seqfeature.qualifiers["locus_tag"][0]
    except (KeyError, IndexError):
        locus_tag = ""
    finally:
        tmrna_ftr.set_locus_tag(locus_tag, delimiter=None)

    tmrna_ftr.set_orientation(seqfeature.strand, "fr_short", True)
    tmrna_ftr.start, tmrna_ftr.stop, tmrna_ftr.parts = parse_coordinates(
                                                                    seqfeature)

    # Coordinate format for GenBank flat file features parsed by Biopython
    # are 0-based half open intervals.
    tmrna_ftr.coordinate_format = "0_half_open"

    tmrna_ftr.set_nucleotide_length(use_seq=True)

    try:
        note = seqfeature.qualifiers["note"][0]
    except (KeyError, IndexError):
        note = ""
    finally:
        tmrna_ftr.note = note

    try:
        gene = seqfeature.qualifiers["gene"][0]
    except (KeyError, IndexError):
        gene = ""
    finally:
        tmrna_ftr.gene = gene

    tmrna_ftr.set_name()
    return tmrna_ftr


def parse_source_seqfeature(seqfeature):
    """Parses a Biopython Source SeqFeature.

    :param seqfeature: Biopython SeqFeature
    :type seqfeature: SeqFeature
    :param genome_id:

        An identifier for the genome in which the seqfeature
        is defined.

    :type genome_id: str
    :returns: A pdm_utils Source object
    :rtype: Source
    """
    src_ftr = source.Source()
    src_ftr.seqfeature = seqfeature
    start, stop, parts = parse_coordinates(seqfeature)
    src_ftr.start = start
    src_ftr.stop = stop

    try:
        src_ftr.organism = str(seqfeature.qualifiers["organism"][0])
    except:
        src_ftr.organism = ""

    try:
        src_ftr.host = str(seqfeature.qualifiers["host"][0])
    except:
        src_ftr.host = ""

    try:
        src_ftr.lab_host = str(seqfeature.qualifiers["lab_host"][0])
    except:
        src_ftr.lab_host = ""

    src_ftr.parse_organism()
    src_ftr.parse_host()
    src_ftr.parse_lab_host()

    return src_ftr


def create_seqfeature_dictionary(seqfeature_list):
    """Create a dictionary of Biopython SeqFeature objects based on their type.

    From a list of all Biopython SeqFeatures derived from a GenBank-formatted
    flat file, create a dictionary of SeqFeatures based on their 'type'
    attribute.

    :param seqfeature_list: List of Biopython SeqFeatures
    :type seqfeature_list: list
    :param genome_id:

        An identifier for the genome in which the seqfeature
        is defined.

    :type genome_id: str
    :returns:

        A dictionary of Biopython SeqFeatures:
        Key: SeqFeature type (source, tRNA, CDS, other)
        Value: SeqFeature

    :rtype: dict
    """

    seqfeature_type_set = set()
    seqfeature_dict = {}
    for seqfeature in seqfeature_list:
        seqfeature_type_set.add(seqfeature.type)
    for type in seqfeature_type_set:
        sublist = []
        for index in range(len(seqfeature_list)):
            seqfeature = seqfeature_list[index]
            if seqfeature.type == type:
                sublist.append(seqfeature)
        seqfeature_dict[type] = sublist
    return seqfeature_dict


# TODO should this function be improved? Maybe create a more basic
# parse_flat_file() function - the only parameter is the seqrecord, and
# there is only minimal parsing and data processing.
# Then the parse_genome_data() function calls parse_flat_file(),
# and processes some data in specific ways.
def parse_genome_data(seqrecord, filepath=pathlib.Path(),
                      translation_table=11, genome_id_field="_organism_name",
                      gnm_type="", host_genus_field="_organism_host_genus"):
    """Parse data from a Biopython SeqRecord object into a Genome object.

    All Source, CDS, tRNA, and tmRNA features are parsed into their
    associated Source, Cds, Trna, and Tmrna objects.

    :param seqrecord: A Biopython SeqRecord object.
    :type seqrecord: SeqRecord
    :param filepath: A filename associated with the returned Genome object.
    :type filepath: Path
    :param translation_table:
        The applicable translation table for the genome's CDS features.
    :type translation_table: int
    :param genome_id_field:
        The SeqRecord attribute from which the unique genome
        identifier/name is stored.
    :type genome_id_field: str
    :param host_genus_field:
        The SeqRecord attribute from which the unique host genus
        identifier/name is stored.
    :type host_genus_field: str
    :param gnm_type: Identifier for the type of genome.
    :type gnm_type: str
    :returns: A pdm_utils Genome object.
    :rtype: Genome
    """

    # Keep track of the file from which the record is derived.
    gnm = genome.Genome()
    gnm.set_filename(filepath)
    gnm.type = gnm_type

    try:
        gnm.organism = seqrecord.annotations["organism"]
    except:
        gnm.organism = ""
    finally:
        # Identifies host and phage name from organism field.
        gnm.parse_organism()

    try:
        # Since accessions are stored in a list, there may be more than
        # one accessions associated with this file.
        # The first accession in the list is assumed to be the most recent.
        accession = seqrecord.annotations["accessions"][0]
    except:
        accession = ""
    finally:
        gnm.set_accession(accession)

    try:
        gnm.description = seqrecord.description
        # It appears that if description is not present, Biopython
        # auto-populates this attribute as "<unknown description>"
        if gnm.description == "<unknown description>":
            gnm.description = ""
    except:
        gnm.description = ""
    finally:
        # Identifies host and phage name from description field.
        gnm.parse_description()

    try:
        gnm.source = seqrecord.annotations["source"]
    except:
        gnm.source = ""
    finally:
        # Identifies host and phage name from record source field.
        gnm.parse_source()

    try:
        # The retrieved authors can be stored in multiple Reference elements.
        refs = seqrecord.annotations["references"]
        authors_list = []
        for ref in refs:
            # Note: Reference objects are instantiated with an empty
            # authors attribute. So if no authors are present in a Reference,
            # it will still concatenate an empty string, resulting in an
            # author_string = ";;;" etc. So only add the authors info if
            # it is not an empty string.
            if ref.authors != "":
                authors_list.append(ref.authors)
        authors_string = ";".join(authors_list)
        gnm.authors = authors_string
    except:
        gnm.authors = ""

    # Biopython requires the parsed record contains a sequence, so
    # no need to test whether the seq attribute is present or not.
    # Nucleotide sequence, length, and % GC.
    gnm.set_sequence(seqrecord.seq)

    try:
        date = seqrecord.annotations["date"]
        gnm.date = datetime.strptime(date, "%d-%b-%Y")
    except:
        gnm.date = constants.EMPTY_DATE

    # # Now that record fields are parsed, set the genome name, id,
    # # and host_genus.
    if genome_id_field != "":
        gnm.name = getattr(gnm, genome_id_field)
        gnm.set_id(value=gnm.name)
    else:
        # The seqrecord name and id are used if genome_id_field is empty.
        try:
            gnm.name = seqrecord.name
            # It appears that if name is not present, Biopython auto-populates
            # this attribute as "<unknown name>"
            if gnm.name == "<unknown name>":
                gnm.name = ""
        except:
            gnm.name = ""

        try:
            gnm.id = seqrecord.id
            # It appears that if id is not present, Biopython auto-populates
            # this attribute as "<unknown id>"
            if gnm.id == "<unknown id>":
                gnm.id = ""
        except:
            gnm.id = ""

    gnm.set_host_genus(attribute=host_genus_field)

    # Create lists of parsed features.
    # Note: Biopython instantiates the features attribute with
    # an empty list, so no need to test if features attribute is
    # present or not.
    seqfeature_dict = create_seqfeature_dictionary(seqrecord.features)

    cds_list = []
    if "CDS" in seqfeature_dict.keys():
        for seqfeature in seqfeature_dict["CDS"]:
            cds_ftr = parse_cds_seqfeature(seqfeature)
            cds_ftr.genome_id = gnm.id
            cds_ftr.genome_length = gnm.length
            cds_ftr.set_nucleotide_sequence(parent_genome_seq=gnm.seq)
            cds_list.append(cds_ftr)

    source_list = []
    if "source" in seqfeature_dict.keys():
        for seqfeature in seqfeature_dict["source"]:
            src_ftr = parse_source_seqfeature(seqfeature)
            src_ftr.genome_id = gnm.id
            source_list.append(src_ftr)

    trna_list = []
    if "tRNA" in seqfeature_dict.keys():
        for seqfeature in seqfeature_dict["tRNA"]:
            trna_ftr = parse_trna_seqfeature(seqfeature)
            trna_ftr.genome_id = gnm.id
            trna_ftr.genome_length = gnm.length
            trna_ftr.set_nucleotide_sequence(parent_genome_seq=gnm.seq)
            trna_ftr.set_nucleotide_length(use_seq=True)
            trna_ftr.parse_amino_acid()
            trna_ftr.parse_anticodon()
            trna_list.append(trna_ftr)

    tmrna_list = []
    if "tmRNA" in seqfeature_dict.keys():
        for seqfeature in seqfeature_dict["tmRNA"]:
            tmrna_ftr = parse_tmrna_seqfeature(seqfeature)
            tmrna_ftr.genome_id = gnm.id
            tmrna_ftr.genome_length = gnm.length
            tmrna_ftr.set_nucleotide_sequence(parent_genome_seq=gnm.seq)
            tmrna_ftr.set_nucleotide_length(use_seq=True)
            tmrna_ftr.parse_peptide_tag()
            tmrna_ftr.run_aragorn()
            tmrna_list.append(tmrna_ftr)

    gnm.translation_table = translation_table
    gnm.set_cds_features(cds_list)
    gnm.set_source_features(source_list)
    gnm.set_trna_features(trna_list)
    gnm.set_tmrna_features(tmrna_list)

    # The feature.id is constructed from the Genome.id and the feature order.
    gnm.set_feature_ids(use_type=True, use_cds=True)
    gnm.set_feature_ids(use_type=True, use_source=True)
    gnm.set_feature_ids(use_type=True, use_trna=True)
    gnm.set_feature_ids(use_type=True, use_tmrna=True)
    return gnm


def create_fasta_seqrecord(header, sequence_string):
    """Create a fasta-formatted Biopython SeqRecord object.

    :param header: Description of the sequence.
    :type header: str
    :param sequence_string: Nucleotide sequence.
    :type sequence_string: str
    :returns: Biopython SeqRecord containing the nucleotide sequence.
    :rtype: SeqRecord
    """
    seq = Seq(sequence_string, alphabet=IUPAC.unambiguous_dna)
    seqrecord = SeqRecord(seq, description=header)
    return seqrecord


# Needs unittests, however:
# Seqfeature retrieval and generation is clunky probably requires some
# over arching seqfeature generation.
# May delay unittests until structure is revamped
def genome_to_seqrecord(phage_genome):
    """Creates a SeqRecord object from a pdm_utils Genome object.

    :param phage_genome: A pdm_utils Genome object.
    :type phage_genome: Genome
    :returns: A BioPython SeqRecord object
    :rtype: SeqRecord
    """
    assert phage_genome is not None, \
        "Genome object passed is None and not initialized"
    try:
        record = SeqRecord(phage_genome.seq)
        record.seq.alphabet = IUPAC.IUPACAmbiguousDNA()
    except AttributeError:
        print("Genome object failed to be converted to SeqRecord.",
              "Genome valid attribute 'seq' is required to",
              "convert to SeqRecord object.")
        raise
    record.name = phage_genome.name
    if phage_genome.accession == "" or phage_genome.accession is None:
        record.id = "".join(["DRAFT_", phage_genome.name])
    else:
        record.id = phage_genome.accession
    record.features = get_genome_seqrecord_features(phage_genome)
    record.description = get_genome_seqrecord_description(phage_genome)
    record.annotations = get_genome_seqrecord_annotations(phage_genome)

    return record


def cds_to_seqrecord(cds, parent_genome, gene_domains=[], desc_type="gb"):
    """Creates a SeqRecord object from a Cds and its parent Genome.

    :param cds: A populated Cds object.
    :type cds: Cds
    :param phage_genome: Populated parent Genome object of the Cds object.
    :param domains: List of domain objects populated with column attributes
    :type domains: list
    :param desc_type: Inteneded format of the CDS SeqRecord description.
    :type desc_type: str
    :returns: Filled Biopython SeqRecord object.
    :rtype: SeqRecord
    """
    record = SeqRecord(cds.translation)
    record.seq.alphabet = IUPAC.IUPACProtein()
    record.name = cds.id
    record.id = cds.id

    cds.set_seqfeature()

    source = f"{parent_genome.host_genus} phage {cds.genome_id}"
    source_feature = cds.create_seqfeature("source", 0,
                                           cds.translation_length, 1)
    source_feature.qualifiers["organism"] = [source]

    record.features = [source_feature]
    record.features.append(cds.create_seqfeature("Protein", 0,
                                                 cds.translation_length, 1))

    cds_feature = cds.create_seqfeature("CDS", 0, cds.translation_length, 1)
    format_cds_seqrecord_CDS_feature(cds_feature, cds, parent_genome)
    record.features.append(cds_feature)

    region_features = get_cds_seqrecord_regions(gene_domains, cds)
    for region_feature in region_features:
        record.features.append(region_feature)

    if desc_type == "fasta":
        description = ""

        product = cds.seqfeature.qualifiers["product"][0]
        if product > "":
            description = "".join([description, f"[product={product}]"])

        cluster = parent_genome.cluster
        if cluster > "":
            description = " ".join([description, f"[cluster={cluster}]"])

        host_genus = parent_genome.host_genus
        if host_genus > "":
            description = " ".join([description, f"[host_genus={host_genus}]"])

        if cds.locus_tag > "":
            description = " ".join([description, f"[locus={cds.locus_tag}]"])

        record.description = description
    elif desc_type == "gb":
        description = f"{source} gp {cds.name}"
        record.description = description

    record.annotations = get_cds_seqrecord_annotations(cds, parent_genome)

    return record


def get_genome_seqrecord_features(phage_genome):
    """Helper function that uses Genome data to populate
    the features SeqRecord atribute

    :param phage_genome:
        Input a Genome object.
    :type phage_genome: genome
    :returns:
        features is a list of SeqFeature objects parsed
        from cds objects
    """

    source_feature = SeqFeature(FeatureLocation(0, phage_genome.length),
                                strand=1, type="source")
    source_feature.qualifiers = OrderedDict()
    source_feature.qualifiers["source"] = (f"{phage_genome.host_genus} phage "
                                           f"{phage_genome.name}")

    features = [source_feature]

    for phage_cds in phage_genome.cds_features:
        phage_cds.set_seqfeature(type="gene")
        features.append(phage_cds.seqfeature)
        phage_cds.set_seqfeature(type="CDS")
        features.append(phage_cds.seqfeature)

    for phage_trna in phage_genome.trna_features:
        phage_trna.set_seqfeature(type="gene")
        features.append(phage_trna.seqfeature)
        phage_trna.set_seqfeature()
        features.append(phage_trna.seqfeature)

    return features


def get_genome_seqrecord_description(phage_genome):
    """Helper function to construct a description SeqRecord attribute.

    :param phage_genome:
        Input a Genome object.
    :type phage_genome: genome
    :returns:
        description is a formatted string parsed
        from genome data
    """
    description = (f"{phage_genome.host_genus} phage {phage_genome.name}"
                   ", complete genome")
    return description


def get_genome_seqrecord_annotations(phage_genome):
    """Helper function that uses Genome data to populate
    the annotations SeqRecord attribute

    :param phage_genome:
        Input a Genome object.
    :type phage_genome: genome
    :returns:
        annotations(dictionary) is a dictionary with
        the formatting of BioPython's SeqRecord
        annotations attribute
    """

    annotations = {"molecule type": "DNA",
                   "topology": "linear",
                   "data_file_division": "PHG",
                   "date": "",
                   "accessions": [],
                   "sequence_version": "1",
                   "keywords": [],
                   "source": "",
                   "organism": "",
                   "taxonomy": [],
                   "comment": ()}
    annotations["date"] = phage_genome.date
    annotations["keywords"] = ["complete_genome"]
    annotations["source"] = (
            f"{phage_genome.host_genus} phage {phage_genome.id}")
    annotations["organism"] = (
            f"{phage_genome.host_genus} phage {phage_genome.name}")
    annotations["taxonomy"].append("Viruses")
    annotations["taxonomy"].append("dsDNA Viruses")
    annotations["taxonomy"].append("Caudovirales")
    annotations["comment"] = get_genome_seqrecord_annotations_comments(
                                                                phage_genome)
    return annotations


def get_genome_seqrecord_annotations_comments(phage_genome):
    """Helper function that uses Genome data to populate
    the comment annotation attribute

    :param phage_genome:
        Input a Genome object.
    :type phage_genome: genome
    :returns:
        cluster_comment, auto_generated_comment
        annotation_status_comment, qc_and_retrieval values
        (tuple) is a tuple with
        the formatting of BioPython's SeqRecord
        annotations comment attribute
    """
    if phage_genome.subcluster == "":
        cluster_comment = "Cluster: {}; Subcluster: None".format(
                                                        phage_genome.cluster)
    else:
        cluster_comment = "Cluster: {}; Subcluster: {}".format(
                                phage_genome.cluster, phage_genome.subcluster)
    auto_generated_comment = (
            "Auto-generated genome record from the MySQL database")
    annotation_status_comment = (
            "Annotation Status: {}; Annotation Author: {}".format(
                                             phage_genome.annotation_status,
                                             phage_genome.annotation_author))
    retrieval_value = (
            "RetrieveRecord: {}".format(phage_genome.retrieve_record))

    return (cluster_comment, auto_generated_comment,
            annotation_status_comment, retrieval_value)


def get_cds_seqrecord_regions(gene_domains, cds):
    region_features = []
    for gene_domain in gene_domains:
        region_feature = cds.create_seqfeature("Region",
                                               gene_domain["QueryStart"],
                                               gene_domain["QueryEnd"], 1)
        region_feature.qualifiers["region_name"] = [gene_domain["Name"]]

        description = gene_domain["Description"]
        if description is None:
            description = ""
        else:
            description = description.decode("utf-8")
        region_feature.qualifiers["note"] = [description]

        region_feature.qualifiers["db_xref"] = ["CDD:"
                                                f"{gene_domain['DomainID']}"]
        region_features.append(region_feature)

    return region_features


def format_cds_seqrecord_CDS_feature(cds_feature, cds, parent_genome):
    cds_feature.qualifiers.pop("codon_start")
    cds_feature.qualifiers.pop("product")
    cds_feature.qualifiers.pop("translation")

    if cds.coordinate_format == "0_half_open":
        start = cds.seqfeature.location.start + 1
    elif cds.coordinate_format == "1_closed":
        start = cds.seqfeature.location.start
    else:
        start = cds.seqfeature.location.start

    coded_by = (f"{parent_genome.accession}:"
                f"{start}..""{cds.seqfeature.location.end}")
    if cds.seqfeature.strand == -1:
        coded_by = f"complement({coded_by})"

    cds_feature.qualifiers["coded by"] = [coded_by]


def get_cds_seqrecord_annotations(cds, parent_genome):
    """Function that creates a Cds SeqRecord annotations attribute dict.
    :param cds: A populated Cds object.
    :type cds: Cds
    :param phage_genome: Populated parent Genome object of the Cds object.
    :type phage_genome: Genome
    :returns: Formatted SeqRecord annotations dictionary.
    :rtype: dict{str}
    """
    annotations = {"topology": "linear",
                   "data_file_division": "PHG",
                   "date": "",
                   "accessions": [],
                   "sequence_version": "",
                   "keywords": [],
                   "source": "",
                   "organism": "",
                   "taxonomy": [],
                   "comment": ()}

    annotations["date"] = parent_genome.date
    annotations["organism"] = (f"{parent_genome.host_genus} phage "
                               f"{cds.genome_id}")
    annotations["source"] = f"Accession {parent_genome.accession}"

    annotations["taxonomy"].append("Viruses")
    annotations["taxonomy"].append("dsDNA Viruses")
    annotations["taxonomy"].append("Caudovirales")

    annotations["comment"] = get_cds_seqrecord_annotations_comments(cds)

    return annotations


def get_cds_seqrecord_annotations_comments(cds):
    """Function that creates a Cds SeqRecord comments attribute tuple.

    :param cds:
    :type cds:
    """
    pham_comment = f"Pham: {cds.pham_id}"
    auto_generated_comment = "Auto-generated CDS record from a MySQL database"

    return (pham_comment, auto_generated_comment)


def sort_seqrecord_features(seqrecord):
    """Function that sorts and processes the seqfeature objects of a seqrecord.

    :param seqrecord: Phage genome Biopython seqrecord object
    :type seqrecord: SeqRecord
    """

    try:
        def _sorting_key(seqfeature): return seqfeature.location.start
        seqrecord.features.sort(key=_sorting_key)
    except:
        if seqrecord is None:
            raise TypeError
        print("Genome seqrecord features unable to be sorted")
        pass



"""Pipeline to automate product annotation resubmissions to GenBank. """
import argparse
import logging
import sys
import time
from datetime import date
from pathlib import Path

import pdm_utils         # to get the version number
from pdm_utils.functions import configfile
from pdm_utils.functions import fileio
from pdm_utils.functions import flat_files
from pdm_utils.functions import ncbi
from pdm_utils.functions import pipelines_basic
from pdm_utils.functions import querying
from pdm_utils.functions import mysqldb

# -----------------------------------------------------------------------------
# GLOBAL VARIABLES

DEFAULT_FOLDER_NAME = (f"{time.strftime('%Y%m%d')}_revise")

BASE_CONDITIONALS = ("phage.Status = final AND "
                     "phage.AnnotationAuthor = 1 AND"
                     "phage.RetrieveRecord = 1")

REVISE_PIPELINES = ["local", "remote"]

CURATION_NAME = "revise.csv"

CURATION_HEADER = ["Phage", "Accession Number", "Locus Tag",
                   "Start", "Stop", "Product"]
TICKET_HEADER = ["table", "field", "value", "key_name", "key_value"]

FIVE_COLUMN_TABLE_HEADER = []

REVISION_COLUMNS = ["phage.PhageID", "phage.Accession", "gene.LocusTag",
                    "gene.Start", "gene.Stop", "gene.Notes", "gene.GeneID"]

LOCAL_INPUT_FILE_TYPES = ["function_report", "csv"]
LOCAL_OUTPUT_FILE_TYPES = ["ticket", "p_curation"]

REMOTE_OUTPUT_FILE_TYPES = ["tbl", "p_curation"]

INPUT_FILE_KEYS = {"function_report": {"data_key": "Pham",
                                       "filter_key": "gene.PhamID"},
                   "csv": {"data_key": "GeneID",
                           "filter_key": "gene.GeneID"}
                   }

PHAGE_QUERY = "SELECT * FROM phage"
GENE_QUERY = "SELECT * FROM gene"
TRNA_QUERY = "SELECT * FROM trna"
TMRNA_QUERY = "SELECT * FROM tmrna"


logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
MAIN_LOG_FILE = "REVISE_LOG.log"
VERSION = pdm_utils.__version__
CURRENT_DATE = date.today().strftime("%Y%m%d")
# MAIN FUNCTIONS
# -----------------------------------------------------------------------------


def main(unparsed_args_list):
    """Uses parsed args to run the entirety of the revise pipeline.

    :param unparsed_args_list: Input a list of command line args.
    :type unparsed_args_list: list[str]
    """
    args = parse_revise(unparsed_args_list)

    config = configfile.build_complete_config(args.config_file)

    alchemist = pipelines_basic.build_alchemist(args.database, config=config)

    if args.pipeline == "local":
        execute_local_revise(alchemist, args.revisions_file,
                             folder_path=args.folder_path,
                             folder_name=args.folder_name,
                             config=config, input_type=args.input_type,
                             output_type=args.output_type,
                             filters=args.filters, groups=args.groups,
                             verbose=args.verbose, force=args.force,
                             production=args.production)
    elif args.pipeline == "remote":
        values = pipelines_basic.parse_value_input(args.input)
        execute_remote_revise(alchemist, folder_path=args.folder_path,
                              folder_name=args.folder_name, config=config,
                              values=values, filters=args.filters,
                              groups=args.groups, verbose=args.verbose,
                              output_type=args.output_type, force=args.force)


def parse_revise(unparsed_args_list):
    """Parses revise arguments and stores them with an argparse object.

    :param unparsed_args_list: Input a list of command line args.
    :type unparsed_args_list: list[str]
    :returns: ArgParse module parsed args.
    """
    DATABASE_HELP = """
        Name of the MySQL database to export from.
        """
    PIPELINE_HELP = """Pipeline of how to retrieve information to submit to
        GenBank
        """

    REMOTE_HELP = """Revise pipeline that accepts changes to submit to GenBank
        based on differences between data in the database and data retrieved
        from GenBank.
        """
    LOCAL_HELP = """Revise pipeline that accepts changes to submit to GenBank
        based on differences between data received from a file and the MySQL
        database.
        """

    REVISIONS_FILE_HELP = """
        Selection input option that imports values from a csv file.
            Follow selection argument with path to the
            csv file containing the names of each genome in the first column.
        """
    INPUT_TYPE_HELP = """
        Revision option that selects the input file type.
            Follow selection argument with a supported file type.
        """
    OUTPUT_TYPE_HELP = """
        Revision option that selects the output_file_type.
            Follow selection argument with a supported file type.
        """
    PRODUCTION_HELP = """
        Revise option to toggle additional filters to support production-level
        revision.
        """

    IMPORT_FILE_HELP = """
        Selection input option that imports values from a csv file.
            Follow selection argument with path to the
            csv file containing the names of each genome in the first column.
        """
    SINGLE_GENOMES_HELP = """
        Selection input option that imports values from cmd line input.
            Follow selection argument with space separated
            names of genomes in the database.
        """

    FORCE_HELP = """
        Revise option that aggresively creates and overwrites directories.
        """
    CONFIG_FILE_HELP = """
        Revise option that enables use of a config file for sourcing
        credentials
            Follow selection argument with the path to the config file
            specifying MySQL and NCBI credentials.
        """
    VERBOSE_HELP = """
        Revise option that enables progress print statements.
        """
    FOLDER_PATH_HELP = """
        Revise option to change the path
        of the directory where the exported files are stored.
            Follow selection argument with the path to the
            desired directory.
        """
    FOLDER_NAME_HELP = """
        Revise option to change the name
        of the directory where the exported files are stored.
            Follow selection argument with the desired name.
        """

    FILTERS_HELP = """
        Data filtering option that filters data by the inputted expressions.
            Follow selection argument with formatted filter expression:
                {Table}.{Column}={Value}
        """
    GROUP_BY_HELP = """
        Data selection option that groups data by the inputted columns.
            Follow selection argument with formatted column expressions:
                {Table}.{Column}={Value}
        """

    parser = argparse.ArgumentParser()

    parser.add_argument("database", type=str,  help=DATABASE_HELP)

    subparsers = parser.add_subparsers(dest="pipeline", required=True,
                                       help=PIPELINE_HELP)

    local_parser = subparsers.add_parser("local", help=LOCAL_HELP)
    remote_parser = subparsers.add_parser("remote", help=REMOTE_HELP)

    local_parser.add_argument("revisions_file", help=REVISIONS_FILE_HELP,
                              type=pipelines_basic.convert_file_path)

    local_parser.add_argument("-it", "--input_type", help=INPUT_TYPE_HELP,
                              choices=LOCAL_INPUT_FILE_TYPES)
    local_parser.add_argument("-ft", "--output_type", help=OUTPUT_TYPE_HELP,
                              choices=LOCAL_OUTPUT_FILE_TYPES)
    local_parser.add_argument("-p", "--production", help=PRODUCTION_HELP,
                              action="store_true")

    remote_parser.add_argument("-if", "--import_file", dest="input",
                               type=pipelines_basic.convert_file_path,
                               help=IMPORT_FILE_HELP)
    remote_parser.add_argument("-in", "--import_names", nargs="*",
                               dest="input", help=SINGLE_GENOMES_HELP)

    remote_parser.add_argument("-ft", "--output_type", help=OUTPUT_TYPE_HELP,
                               choices=REMOTE_OUTPUT_FILE_TYPES)

    for subparser in [local_parser, remote_parser]:
        subparser.add_argument("-f", "--force", action="store_true",
                               help=FORCE_HELP)
        subparser.add_argument("-c", "--config_file", help=CONFIG_FILE_HELP,
                               type=pipelines_basic.convert_file_path)
        subparser.add_argument("-m", "--folder_name", type=str,
                               help=FOLDER_NAME_HELP)
        subparser.add_argument("-o", "--folder_path", type=Path,
                               help=FOLDER_PATH_HELP)
        subparser.add_argument("-v", "--verbose", action="store_true",
                               help=VERBOSE_HELP)

        subparser.add_argument("-w", "--where", nargs="?", dest="filters",
                               help=FILTERS_HELP)
        subparser.add_argument("-g", "--group_by", nargs="*", dest="groups",
                               help=GROUP_BY_HELP)

        subparser.set_defaults(folder_name=DEFAULT_FOLDER_NAME,
                               folder_path=None, config_file=None, input=[],
                               input_type="function_report",
                               output_type="p_curation",
                               filters="", groups=[], verbose=False)

    parsed_args = parser.parse_args(unparsed_args_list[2:])
    return parsed_args


def execute_local_revise(alchemist, revisions_file_path, folder_path=None,
                         folder_name=DEFAULT_FOLDER_NAME, config=None,
                         input_type="function_report",
                         output_type="p_curation", production=False,
                         filters="", groups=[],
                         force=False, verbose=False):
    """Executes the entirety of the genbank local revise pipeline.

    :param alchemist: A connected and fully built AlchemyHandler object.
    :type alchemist: AlchemyHandler
    :param revisions_data_dicts: Data dictionaries containing pham/notes data.
    :type revisions_data_dicts: list[dict]
    :param folder_path: Path to a valid dir for new dir creation.
    :type folder_path: Path
    :param folder_name: A name for the export folder.
    :type folder_name: str
    :param input_type: Specifies the file format of the input file
    :type input_type: str
    :param output_type: Specifies the file format of the outputted file
    :type output_type: str
    :param production: Toggles additional filters for production-level revision
    :type production: bool
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    :param force: A boolean to toggle aggresive building of directories.
    :type force: bool
    """
    keys = INPUT_FILE_KEYS.get(input_type)
    if keys is None:
        raise ValueError(f"Revision input type {input_type} is not supported.")

    revisions_data_dicts = fileio.retrieve_data_dict(revisions_file_path)

    values = []
    for data_dict in revisions_data_dicts:
        values.append(data_dict[keys['data_key']])

    db_filter = pipelines_basic.build_filter(alchemist, keys['filter_key'],
                                             filters, values=values,
                                             verbose=verbose)

    if production:
        db_filter.add(BASE_CONDITIONALS)

    revise_columns = db_filter.get_columns(REVISION_COLUMNS)

    if verbose:
        print("Creating export folder...")
    export_path = pipelines_basic.create_working_path(folder_path, folder_name,
                                                      force=force)

    conditionals_map = pipelines_basic.build_groups_map(db_filter, export_path,
                                                        force=force,
                                                        groups=groups,
                                                        verbose=verbose)

    if verbose:
        print("Prepared query and path structure, beginning revise export...")

    for mapped_path in conditionals_map.keys():
        conditionals = conditionals_map[mapped_path]

        if input_type == "function_report":
            export_dicts = use_function_report_data(
                                        db_filter, revisions_data_dicts,
                                        revise_columns, conditionals,
                                        verbose=verbose)
        elif input_type == "csv":
            export_dicts = use_csv_data(db_filter, revisions_data_dicts,
                                        revise_columns, conditionals,
                                        verbose=verbose)

        if not export_dicts:
            if verbose:
                print(f"'{mapped_path.name}' data selected does not require "
                      "revision; no file exported...")

            continue

        pipelines_basic.create_working_dir(mapped_path, force=force)

        write_revise_file(export_dicts, mapped_path, file_format=output_type,
                          verbose=verbose)


# TODO Owen unittest
def execute_remote_revise(alchemist, folder_path=None,
                          folder_name=DEFAULT_FOLDER_NAME, config=None,
                          output_type="p_curation", values=None,
                          filters="", groups=[], verbose=False, force=False):
    ncbi_creds = {}
    if config is not None:
        ncbi_creds = config["ncbi"]

    db_filter = pipelines_basic.build_filter(alchemist, "phage", filters,
                                             values=values, verbose=verbose)
    db_filter.add(BASE_CONDITIONALS)

    revise_path = pipelines_basic.create_working_path(folder_path, folder_name,
                                                      force=force)

    conditionals_map = pipelines_basic.build_groups_map(
                                    db_filter, revise_path, groups=groups,
                                    verbose=verbose, force=force)

    values = db_filter.values
    for mapped_path in conditionals_map.keys():
        db_filter.reset()
        db_filter.values = values

        conditionals = conditionals_map[mapped_path]
        db_filter.values = db_filter.build_values(where=conditionals)

        if db_filter.hits() == 0:
            print(f"No database entries received for '{mapped_path}'.")

        pipelines_basic.create_working_dir(mapped_path, force=force)
        build_revise_log_file(mapped_path)

        logger.info(f"pdm_utils version: {VERSION}")
        logger.info(f"Revise run date: {CURRENT_DATE}")
        logger.info(f"Connected to database: {alchemist.database}")

        accession_data = db_filter.select(["phage.PhageID", "phage.Accession"])

        acc_id_dict = {}
        for data_dict in accession_data:
            accession = data_dict["Accession"]
            if not (accession is None or accession == ""):
                acc_id_dict[accession] = data_dict["PhageID"]

        tbl_records = get_tbl_records(acc_id_dict, ncbi_cred_dict=ncbi_creds)

        validated_phages = []
        for tbl_record in tbl_records:
            validated_phages.append(tbl_record.name)

        id_record_map = build_id_record_map(alchemist, validated_phages)

        if output_type == "tbl":
            revised_records = revise_seqrecords(id_record_map, tbl_records,
                                                verbose=verbose)

            if not revised_records:
                print("No discrepancies detected between "
                      f"local data and GenBank data for '{mapped_path}'.")
                continue

        elif output_type == "p_curation":
            curation_data_dicts = find_product_discrepancies(id_record_map,
                                                             tbl_records,
                                                             verbose=verbose)

            if not curation_data_dicts:
                print("No discrepancies detected between "
                      f"local data and GenBank data for '{mapped_path}'.")
                continue

        if output_type == "tbl":
            fileio.write_feature_table(revised_records, mapped_path,
                                       verbose=verbose)
        elif output_type == "p_curation":
            file_path = mapped_path.joinpath("revise.csv")
            fileio.export_data_dict(curation_data_dicts, file_path,
                                    CURATION_HEADER, include_headers=True)

# LOCAL REVISE HELPER FUNCTIONS
# -----------------------------------------------------------------------------


def use_function_report_data(db_filter, data_dicts, columns, conditionals,
                             verbose=False):
    """Reads in FunctionReport data and pairs it with existing data.

    :param db_filter: A connected and fully built Filter object.
    :type db_filter: Filter
    :param data_dicts: List of data dictionaries from a FunctionReport file.
    :type data_dicts: list[dict]
    :param columns: List of SQLAlchemy Columns to retrieve data for.
    :type columns: list[Column]
    :param conditionals: List of SQLAlchemy BinaryExpressions to filter with.
    :type conditionals: List[BinaryExpression]
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    """
    if verbose:
        print("Retreiving feature data using pham function report...")

    export_dicts = []
    for data_dict in data_dicts:
        final_call = data_dict["Final Call"]
        if final_call.lower() == "hypothetical protein":
            final_call = ""
        conditionals.append(querying.build_where_clause(
                            db_filter.graph, f"gene.Notes!='{final_call}'"))

        query = querying.build_select(db_filter.graph, columns,
                                      where=conditionals)

        results = querying.execute(db_filter.engine, query,
                                   in_column=db_filter.key,
                                   values=[data_dict["Pham"]])

        for result in results:
            if (not result["Accession"]) or (not result["LocusTag"]):
                continue
            result["Notes"] = data_dict["Final Call"]
            result["Start"] = result["Start"] + 1
            export_dicts.append(result)

    return export_dicts


# TODO Unittest
def use_csv_data(db_filter, data_dicts, columns, conditionals, verbose=False):
    """Reads in gene table csv data and pairs it with existing data.

    :param db_filter: A connected and fully built Filter object.
    :type db_filter: Filter
    :param data_dicts: List of data dictionaries from a FunctionReport file.
    :type data_dicts: list[dict]
    :param columns: List of SQLAlchemy Columns to retrieve data for.
    :type columns: list[Column]
    :param conditionals: List of SQLAlchemy BinaryExpressions to filter with.
    :type conditionals: List[BinaryExpression]
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    """
    if verbose:
        print("Retrieving feauture data using gene table csv...")

    query = querying.build_select(db_filter.graph, columns, where=conditionals)
    results = querying.execute(db_filter.engine, query,
                               in_column=db_filter.key,
                               values=db_filter.values)

    results_dict = {}
    for result in results:
        results_dict['GeneID'] = result

    export_dicts = []
    for data_dict in data_dicts:
        result_dict = results_dict.get(data_dict['GeneID'])
        if result_dict is None:
            continue
        elif result_dict["Notes"].decode("utf-8") != data_dict["Notes"]:
            result_dict["Notes"] = data_dict["Notes"]
            export_dicts.append(result_dict)

    return export_dicts


def write_revise_file(data_dicts, output_path, file_format="p_curation",
                      file_name=CURATION_NAME, verbose=False):
    """Writes a revision csv in the desired file format with necessary changes.

    :param data_dicts: List of data dictionaries to convert to curation format.
    :type data_dicts: list[dict]
    :param output_path: Path to a dir for file creation.
    :type output_path: Path
    :param file_format: Format of the csv to be written.
    :type file_format: str
    :param file_name: Name of the file to write curation data to.
    :type file_name: str
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    """
    data_dicts = sorted(data_dicts, key=lambda data_dict: data_dict["PhageID"])

    if file_format == "p_curation":
        format_data = format_curation_data
    elif file_format == "ticket":
        format_data = format_update_ticket_data
    else:
        raise ValueError(f"File format '{file_format}' is not "
                         "recognized for revise output.")

    for d in data_dicts:
        format_data(d)

    if verbose:
        print(f"Writing {file_name} in {output_path.name}...")
    file_path = output_path.joinpath(file_name)

    if file_format == "p_curation":
        header = CURATION_HEADER
    elif file_format == "ticket":
        header = TICKET_HEADER

    fileio.export_data_dict(data_dicts, file_path, header,
                            include_headers=True)


def format_curation_data(row_dict):
    """Function to format revise dictionary keys to gb curation format.

    :param row_dict: Data dictionary for a revise file.
    :type row_dict: dict
    :param product: Gene product to append to the revise data dictionary.
    :type product: str
    """
    row_dict.pop("GeneID")

    row_dict["Phage"] = row_dict.pop("PhageID")
    row_dict["Accession Number"] = row_dict.pop("Accession")
    row_dict["Locus Tag"] = row_dict.pop("LocusTag")
    row_dict["Product"] = row_dict.pop("Notes")


def format_update_ticket_data(row_dict):
    """Function to format revise dictionary keys to update ticket format.

    :param row_dict: Data dictionary for a revise file.
    :type row_dict: dict
    :param product: Gene product to append to the revise data dictionary.
    :type product: str
    """
    row_dict.pop("PhageID")
    row_dict.pop("Accession")
    row_dict.pop("LocusTag")
    row_dict.pop("Start")
    row_dict.pop("Stop")

    product = row_dict.pop("Notes")
    if product.lower() == "hypothetical protein":
        product = ""

    row_dict["table"] = "gene"
    row_dict["field"] = "Notes"
    row_dict["value"] = product
    row_dict["key_name"] = "GeneID"
    row_dict["key_value"] = row_dict.pop("GeneID")


# REMOTE REVISE HELPER FUNCTIONS
# -----------------------------------------------------------------------------
def build_revise_log_file(working_dir):
    log_file = working_dir.joinpath(MAIN_LOG_FILE)
    logging.basicConfig(filename=log_file, filemode="w",
                        level=logging.DEBUG,
                        format="pdm_utils revise: %(levelname)s: %(message)s")


def get_tbl_records(acc_id_dict, ncbi_cred_dict={}):
    filehandle = ncbi.get_verified_data_handle(acc_id_dict,
                                               ncbi_cred_dict=ncbi_cred_dict,
                                               file_type="tbl")
    tbl_records = []
    if filehandle is None:
        return tbl_records

    record_gen = fileio.parse_feature_table(filehandle)

    for tbl_record in record_gen:
        phage_name = acc_id_dict[tbl_record.id]
        tbl_record.name = phage_name
        flat_files.sort_seqrecord_features(tbl_record)
        tbl_records.append(tbl_record)

    return tbl_records


def build_id_record_map(alchemist, phageids):
    id_record_map = {}
    if not phageids:
        return id_record_map

    genomes = mysqldb.parse_genome_data(alchemist.engine,
                                        phage_id_list=phageids,
                                        phage_query=PHAGE_QUERY,
                                        gene_query=GENE_QUERY,
                                        trna_query=TRNA_QUERY,
                                        tmrna_query=TMRNA_QUERY)

    for genome in genomes:
        record = flat_files.genome_to_seqrecord(genome)
        id_record_map[record.id] = record

    return id_record_map


def revise_seqrecords(source_id_record_map, target_records, verbose=False):
    revised_records = []

    for target_record in target_records:
        source_record = source_id_record_map.get(target_record.id)

        if source_record is None:
            raise Exception

        edits = revise_seqrecord(target_record, source_record, verbose=verbose)
        if len(edits) > 1:
            flat_files.sort_seqrecord_features(target_record)
            revised_records.append(target_record)

        for edit_msg in edits:
            logger.info(edit_msg)

    return revised_records


def find_product_discrepancies(source_id_record_map, target_records,
                               verbose=False):
    total_product_discrepancies = []

    for target_record in target_records:
        source_record = source_id_record_map.get(target_record.id)

        if source_record is None:
            raise Exception

        record_discrepancies = curate_record_product_discrepancies(
                                target_record, source_record, verbose=verbose)
        total_product_discrepancies += record_discrepancies

    return total_product_discrepancies


def curate_record_product_discrepancies(target_record, template_record,
                                        verbose=False):
    record_discrepancies = []
    target_feature_map = create_feature_map(target_record)
    template_feature_map = create_feature_map(template_record)

    for key, target_feature_dict in target_feature_map.items():
        target_cds_feature = target_feature_dict.get("CDS")
        if target_cds_feature is None:
            continue

        temp_feature_dict = template_feature_map.get(key)
        if target_cds_feature is None:
            continue

        temp_cds_feature = temp_feature_dict.get("CDS")
        if temp_cds_feature is None:
            continue

        target_products = target_cds_feature.qualifiers["product"]
        template_products = temp_cds_feature.qualifiers["product"]
        if target_products[0] != template_products[0]:
            discrepancy = {"Phage": template_record.name,
                           "Accession Number": template_record.id,
                           "Locus Tag": key,
                           "Start": target_cds_feature.location.start+1,
                           "Stop": target_cds_feature.location.end,
                           "Product": template_products[0]}

            record_discrepancies.append(discrepancy)

    return record_discrepancies


def revise_seqrecord(target_record, template_record, verbose=False):
    """Function to edit a target record based on data from a template record.

    :param target_record: SeqRecord object to be changed based on the template.
    :type target_record: SeqRecord
    :param template_record: SeqRecord object to be used as a source of data.
    :type template_record: SeqRecord
    """
    target_feature_map = create_feature_map(target_record)
    template_feature_map = create_feature_map(template_record)

    if target_record.id != template_record.id:
        abort_msg = ("Invalid accession detected in the data for "
                     f"{template_record.name}, aborting evaluation...")
        return [abort_msg]

    record_msg = f"Evaluating {target_record.name} from GenBank..."
    if verbose:
        print(record_msg)

    edits = [record_msg]

    for key, temp_feature_dict in template_feature_map.items():
        if key == "" or key is None:
            abort_msg = ("Invalid Locus Tag detected in the local data for "
                         f"{template_record.name}, aborting evaluation...")
            return [abort_msg]

        feature_dict = target_feature_map.get(key)
        if feature_dict is None:
            add_msg = f"\tAdding '{key}' feature(s)..."
            if verbose:
                print(add_msg)
            for feature in temp_feature_dict.values():
                clean_feature(feature)
                target_record.append(feature)
            edits.append(add_msg)
        else:
            if "CDS" in temp_feature_dict.keys():
                temp_cds = feature_dict["CDS"]
                cds = temp_feature_dict.get("CDS")
                if cds is None:
                    add_msg = f"\tAdding '{key}' CDS feature..."
                    if verbose:
                        print(add_msg)

                    clean_feature(temp_cds)
                    target_record.append(temp_cds)
                    edits.append(add_msg)
            if "tRNA" in temp_feature_dict.keys():
                temp_trna = feature_dict["tRNA"]
                trna = temp_feature_dict.get("tRNA")
                if trna is None:
                    add_msg = f"\tAdding '{key}' tRNA feature..."
                    if verbose:
                        print(add_msg)

                    clean_feature(temp_cds)
                    target_record.append(temp_trna)
                    edits.append(add_msg)
            if "tmRNA" in feature_dict.keys():
                pass

    for key, feature_dict in target_feature_map.items():
        if key == "" or key is None:
            abort_msg = ("Invalid Locus Tag detected in the GenBank data for "
                         f"{target_record.name}, aborting evaluation...")
            return [abort_msg]

        temp_feature_dict = template_feature_map.get(key)
        if temp_feature_dict is None:

            remove_msg = f"\tRemoving '{key}' feature(s)..."
            if verbose:
                print(remove_msg)
            edits.append(remove_msg)
            for feature in feature_dict.values():
                target_record.features.remove(feature)
        else:
            gene = feature_dict["gene"]
            temp_gene = temp_feature_dict["gene"]
            gene_edits = revise_gene_feature(gene, temp_gene, key,
                                             verbose=verbose)
            if gene_edits:
                edits = edits + gene_edits

            if "CDS" in feature_dict.keys():
                cds = feature_dict["CDS"]
                temp_cds = temp_feature_dict.get("CDS")
                if temp_cds is None:
                    remove_msg = f"\tRemoving '{key}' CDS feature..."
                    if verbose:
                        print(remove_msg)
                    edits.append(remove_msg)
                    target_record.features.remove(cds)
                else:
                    cds_edits = revise_cds_feature(cds, temp_cds, key,
                                                   verbose=verbose)
                    if cds_edits:
                        edits = edits + cds_edits

            if "tRNA" in feature_dict.keys():
                trna = feature_dict["tRNA"]
                temp_trna = temp_feature_dict.get("tRNA")
                if temp_trna is None:
                    remove_msg = f"\tRemoving '{key}' CDS feature..."
                    if verbose:
                        print(remove_msg)
                    edits.append(remove_msg)
                    target_record.features.remove(trna)
                else:
                    trna_edits = revise_cds_feature(trna, temp_trna, key,
                                                    verbose=verbose)
                    if trna_edits:
                        edits = edits + trna_edits
            if "tmRNA" in feature_dict.keys():
                pass

    return edits


def revise_gene_feature(target_gene, template_gene, locus_tag, verbose=False):
    edits = []

    target_start = target_gene.location.start
    template_start = template_gene.location.start
    target_stop = target_gene.location.end
    template_stop = template_gene.location.end
    if target_start != template_start and target_stop == template_stop:
        start_edit_msg = (f"\tEditing '{locus_tag}' gene feature start from "
                          f"{target_start} to {template_start}...")
        if verbose:
            print(start_edit_msg)

        edits.append(start_edit_msg)
        target_gene.location = template_gene.location

    return edits


def revise_cds_feature(target_cds, template_cds, locus_tag, verbose=False):
    edits = []

    target_start = target_cds.location.start
    template_start = template_cds.location.start
    target_stop = target_cds.location.end
    template_stop = template_cds.location.end
    if target_start != template_start and target_stop == template_stop:
        start_edit_msg = (f"\tEditing '{locus_tag}' CDS feature start from "
                          f"{target_start} to {template_start}...")

        if verbose:
            print(start_edit_msg)

        edits.append(start_edit_msg)
        target_cds.location = template_cds.location

    target_products = target_cds.qualifiers["product"]
    template_products = template_cds.qualifiers["product"]
    if target_products != template_products:
        product_edit_msg = (f"\tEditing '{locus_tag}' CDS feature "
                            f"product from {';'.join(target_products)} "
                            f"to {';'.join(template_products)}")

        if verbose:
            print(product_edit_msg)

        edits.append(product_edit_msg)
        target_cds.qualifiers["product"] = template_products

    return edits


def revise_trna_feature(target_trna, template_trna, locus_tag, verbose=False):
    edits = []

    target_start = target_trna.location.start
    template_start = template_trna.location.start
    target_stop = target_trna.location.end
    template_stop = template_trna.location.end
    if target_start != template_start and target_stop == template_stop:
        start_edit_msg = (f"Editing '{locus_tag}' tRNA feature start...")

        if verbose:
            print(start_edit_msg)

        edits.append(start_edit_msg)
        target_trna.location = template_trna.location

    target_product = target_trna.qualifiers["product"][0]
    template_product = template_trna.qualifiers["product"][0]
    if target_product != template_product:
        product_edit_msg = (f"\tEditing '{locus_tag}' tRNA feature product "
                            f"from {target_product} to {template_product}...")

        if verbose:
            print(product_edit_msg)

        edits.append(product_edit_msg)
        target_trna.qualifiers["product"] = [template_product]

    return edits


def create_feature_map(record):
    """Revise helper function to map all the qualities of one locus tag.

    :param record: SeqRecord object to map gene features for.
    :type record: SeqRecord
    :returns: Returns a dictionary mapping locus_tags to features.
    :rtype: dict
    """
    cds_map = {}
    for feature in record.features:
        if feature.type == "gene":
            locus_tag = feature.qualifiers["locus_tag"][0]
            cds_map[locus_tag] = {"gene": feature}

    for locus, feature_dict in cds_map.items():
        gene_feature = feature_dict["gene"]
        for feature in record.features:
            if feature.location.end == gene_feature.location.end:
                if feature.type == "gene":
                    continue

                feature_dict[feature.type] = feature

    return cds_map


def clean_feature(feature):
    """Revise helper function to format a SeqFeature for Feature Table export.

    :param feature: Biopython SeqFeature object populated with MySQL data.
    :type feature: Biopython
    """
    qualifiers = feature.qualifiers
    temp_qualifiers = {}

    if feature.type == "gene":
        temp_qualifiers["gene"] = [qualifiers.get("gene")]
        temp_qualifiers["locus_tag"] = [qualifiers.get("locus_tag")]
    elif feature.type == "CDS":
        temp_qualifiers["product"] = [qualifiers.get("product")]
        temp_qualifiers["transl_table"] = [qualifiers.get("transl_table")]
        temp_qualifiers["note"] = [qualifiers.get("note")]
    elif feature.type == "tRNA":
        temp_qualifiers["product"] = [qualifiers.get("product")]
        temp_qualifiers["note"] = [qualifiers.get("note")]
    elif feature.type == "tmRNA":
        pass

    feature.qualifiers = temp_qualifiers


if __name__ == "__main__":
    args = sys.argv
    args.insert(0, "")
    main(args)



"""Evaluation mode functions and dictionaries."""

from pdm_utils.functions import basic

# Eval_flag definitions.
EVAL_FLAGS = {
    # Options that impact how data is processed but are not utilized
    # for evaluation of specific parts of a flat file:
    "check_replace": "Should unexpected genome replacements be reported? ",
    "import_locus_tag": "Should CDS feature locus_tags be imported? ",

    # Options that are utilized during the evaluation stage:
    "check_locus_tag": "Should the structure of CDS feature locus_tags be evaluated? ",
    "check_description_tally": "Should the number of descriptions be evaluated? ",
    "check_description_field": "Should CDS descriptions in unexpected fields be reported? ",
    "check_description": "Should unexpected CDS descriptions be reported? ",
    "check_trna": "Should tRNA features be evaluated? ",
    "check_tmrna": "Should tmRNA features be evaluated? ",
    "check_id_typo": "Should genome ID typos be reported? ",
    "check_host_typo": "Should host typos be reported? ",
    "check_author": "Should unexpected authors be reported? ",
    "check_gene": "Should the CDS 'gene' qualifier be evaluated? ",
    "check_seq": "Should the nucleotide sequence be evaluated? ",
    "check_coords": "Should feature duplication be evaluated? "
    }

# Evaluation mode definitions.
EVAL_MODES = {
    "draft": ("Relaxed evaluations for automatically generated draft "
              "genome annotations since data have not been manually reviewed."),
    "final": ("Stringent evaluations for manual genome annotations."),
    "auto": ("Relaxed evaluations for genome annotations "
             "automatically retrieved from GenBank for routine replacement."),
    "misc": ("Relaxed evaluations for genome annotations "
             "that have been generated from external sources."),
    "custom": ("User-defined evaluations for customized import.")
    }

def get_eval_flag_dict(eval_mode):
    """Get a dictionary of evaluation flags.

    :param eval_mode:
        Valid evaluation mode (base, draft, final, auto, misc, custom)
    :type eval_mode: str
    :returns: Dictionary of boolean values.
    :rtype: dict
    """

    # Base dictionary with all flags set to True.
    dict = {}
    for key in EVAL_FLAGS:
        dict[key] = True

    # Auto-annotations.
    if eval_mode == "draft":
        dict["check_locus_tag"] = False
        dict["check_trna"] = False
        dict["check_tmrna"] = False
        dict["import_locus_tag"] = False
        dict["check_id_typo"] = False
        dict["check_host_typo"] = False
        dict["check_author"] = False
        dict["check_description"] = False
        dict["check_coords"] = False

    # Manual annotations.
    elif eval_mode == "final":
        dict["import_locus_tag"] = False

    # SEA-PHAGES GenBank records.
    elif eval_mode == "auto":
        dict["check_locus_tag"] = False
        dict["check_description_field"] = False
        dict["check_replace"] = False
        dict["check_trna"] = False
        dict["check_tmrna"] = False
        dict["check_id_typo"] = False
        dict["check_host_typo"] = False
        dict["check_author"] = False
        dict["check_description"] = False
        dict["check_description_tally"] = False
        dict["check_gene"] = False
        dict["check_coords"] = False

    # Non-SEA-PHAGES GenBank records.
    elif eval_mode == "misc":
        dict["check_locus_tag"] = False
        # TODO below should probably be True, but it causes problems
        # when checking the current genome, GNM2_001, since these are not 'draft'
        # genomes.
        dict["check_replace"] = False
        dict["check_trna"] = False
        dict["check_tmrna"] = False
        dict["check_id_typo"] = False
        dict["check_host_typo"] = False
        dict["check_author"] = False
        dict["check_description"] = False
        dict["check_description_tally"] = False
        dict["check_gene"] = False

    # Custom QC settings. User can select the settings, so it is initialized as
    # a copy of the base eval_mode. The user can provide the
    # customized combination of options.
    elif eval_mode == "custom":
        for key in dict.keys():
            prompt = f"Eval_flag: {key}. {EVAL_FLAGS[key]}"
            response = basic.ask_yes_no(prompt=prompt, response_attempt=3)
            if response is None:
                print("The default setting for this eval_flag will be used.")
            else:
                dict[key] = response

    elif eval_mode == "base":
        pass
    else:
        print("A valid eval_mode has not been selected.")
    return dict



import csv
import os
import textwrap
from pathlib import Path

from Bio import SeqIO
from Bio.SeqFeature import CompoundLocation

from pdm_utils.classes.fileio import FeatureTableParser
from pdm_utils.functions import multithread

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
TBL_EXCLUDED_QUALIFIERS = ["translation"]
TBL_SPECIAL_QUALIFIERS = ["ribosomal_slippage"]


# READING FUNCTIONS
# -----------------------------------------------------------------------------
def retrieve_data_dict(filepath):
    """Open file and retrieve a dictionary of data.

    :param filepath:
        Path to file containing data and column names.
    :type filepath: Path
    :returns:
        A list of elements, where each element is a dictionary
        representing one row of data.
        Each key is a column name and each value is the data
        stored in that field.
    :rtype: list
    """
    data_dicts = []
    with filepath.open(mode='r') as file:
        file_reader = csv.DictReader(file)
        for dict in file_reader:
            data_dicts.append(dict)
    return data_dicts


def read_feature_table(filehandle):
    """Reads a (five-column) feature table and parses the data into a seqrecord.

    :param filepath: Path to the five-column formatted feature table file.
    :type filepath: Path
    :returns: Returns a Biopython SeqRecord object with the table data.
    :rtype: SeqRecord
    """
    parser = FeatureTableParser(filehandle)

    try:
        record = parser.next()
    except:
        record = None

    return record


def parse_feature_table(filehandle):
    """Takes a (five-column) feature table(s) file handle and parses the data.

    :param filehandle: Handle for a five-column formatted feature table file:
    :returns: Returns a feature table file parser generator.
    :rtype: FeatureTableFileParser
    """
    return FeatureTableParser(filehandle)


def reintroduce_fasta_duplicates(ts_to_gs, filepath):
    """Reads a fasta file and reintroduces (rewrittes) duplicate sequences
    guided by an ungapped translation to sequence-id map

    :param filepath: Path to fasta-formatted multiple sequence file
    :type filepath: pathlib.Path
    :param ts_to_gs: Dictionary mapping unique translations to sequence-ids
    :type ts_to_gs: dict
    """
    with filepath.open(mode="r") as filehandle:
        records = []
        for record in SeqIO.parse(filehandle, "fasta"):
            records.append(record)

    gs_to_ts = {}
    for record in records:
        translation = record.seq
        ungapped_translation = translation.ungap(gap="-")

        geneids = ts_to_gs.get(str(ungapped_translation))
        if geneids is not None:
            for geneid in geneids:
                gs_to_ts[geneid] = str(translation)
        else:
            gs_to_ts[record.id] = str(translation)

    write_fasta(gs_to_ts, filepath)


# WRITING FUNCTIONS
# -----------------------------------------------------------------------------
def export_data_dict(data_dicts, file_path, headers, include_headers=False):
    """Save a dictionary of data to file using specified column headers.

    Ensures the output file contains a specified number of columns,
    and it ensures the column headers are exported as well.

    :param data_dicts:
        list of elements, where each element is a dictionary.
    :type data_dicts: list
    :param file_path: Path to file to export data.
    :type file_path: Path
    :param headers:
        List of strings to define the column order in the file.
        If include_headers is selected, the first row of the file
        will contain each string.
    :type headers: list
    :param include_headers:
        Indicates whether the file should contain a
        row of column names derived from the headers parameter.
    :type include_headers: bool
    """

    headers_dict = {}
    for header in headers:
        headers_dict[header] = header
    # with open(file_path, "w") as file_handle:
    with file_path.open("w") as file_handle:
        file_writer = csv.DictWriter(file_handle, headers)
        if include_headers:
            file_writer.writerow(headers_dict)
        for data_dict in data_dicts:
            file_writer.writerow(data_dict)


def write_fasta(ids_seqs, infile_path, name=None):
    """
    Writes the input genes to the indicated file in FASTA multiple
    sequence format (unaligned).
    :param id_seqs: the ids and sequences to be written to file
    :type genes: dict
    :param infile_path: the path of the file to write the genes to
    :type infile: Path
    :type infile: str
    """
    if isinstance(infile_path, str):
        file_handle = open(infile_path, "w")
    elif isinstance(infile_path, Path):
        file_handle = infile_path.open(mode="w")
    else:
        raise TypeError("File path type not supported.")

    if name is not None:
        file_handle.write(f"#{name}\n")

    for id, seq in ids_seqs.items():
        split_seq = textwrap.wrap(seq, 60)
        wrapped_seq = "\n".join(split_seq)
        file_handle.write(f">{id}\n{wrapped_seq}\n\n")

    file_handle.close()


def write_database(alchemist, version, export_path, db_name=None):
    """Output .sql file from the selected database.

    :param alchemist: A connected and fully built AlchemyHandler object.
    :type alchemist: AlchemyHandler
    :param version: Database version information.
    :type version: int
    :param export_path: Path to a valid dir for file creation.
    :type export_path: Path
    """
    if db_name is None:
        db_name = alchemist.database

    sql_path = export_path.joinpath(f"{db_name}.sql")
    os.system(f"mysqldump -u {alchemist.username} -p{alchemist.password} "
              f"--skip-comments {alchemist.database} > {str(sql_path)}")
    version_path = sql_path.with_name(f"{db_name}.version")
    version_path.touch()
    version_path.write_text(f"{version}")


def write_seqrecord(seqrecord, file_path, file_format):
    file_handle = file_path.open(mode="w")

    if isinstance(seqrecord, list):
        for record in seqrecord:
            SeqIO.write(record, file_handle, file_format)
            file_handle.write("\n")
    else:
        SeqIO.write(seqrecord, file_handle, file_format)

    file_handle.close()


def write_feature_table(seqrecord_list, export_path, verbose=False):
    """Outputs files as five_column tab-delimited text files.

    :param seq_record_list: List of populated SeqRecords.
    :type seq_record_list: list[SeqRecord]
    :param export_path: Path to a dir for file creation.
    :type export_path: Path
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    """
    if verbose:
        print("Writing selected data to files...")
    for record in seqrecord_list:
        if verbose:
            print(f"...Writing {record.name}...")
        file_name = f"{record.name}.tbl"
        file_path = export_path.joinpath(file_name)
        file_handle = file_path.open(mode='w')

        accession = record.id
        version = record.annotations.get("sequence_version")
        if version:
            accession = ".".join([accession, version])

        prefix = record.annotations.get("tbl_prefix")
        if not prefix:
            prefix = ""

        file_handle.write(f">Feature {prefix}|{accession}|\n")

        for feature in record.features:
            location = feature.location
            if isinstance(feature.location, CompoundLocation):
                location = feature.location.parts[0]

            if feature.strand == 1:
                start = location.start + 1
                stop = location.end
            elif feature.strand == -1:
                start = location.end
                stop = location.start + 1

            file_handle.write(f"{start}\t{stop}\t{feature.type}\n")

            if isinstance(feature.location, CompoundLocation):
                if len(feature.location.parts) > 1:
                    for location in feature.location.parts[1:]:
                        if feature.strand == 1:
                            start = location.start + 1
                            stop = location.end
                        elif feature.strand == -1:
                            start = location.end
                            stop = location.start + 1

                        file_handle.write(f"{start}\t{stop}\n")

            for key in feature.qualifiers.keys():
                if key in TBL_EXCLUDED_QUALIFIERS:
                    continue
                elif key in TBL_SPECIAL_QUALIFIERS:
                    file_handle.write(f"\t\t\t{key}\n")
                    continue

                qualifier_values = feature.qualifiers[key]
                for value in qualifier_values:
                    file_handle.write(f"\t\t\t{key}\t{value}\n")

        file_handle.write("\n")
        file_handle.close()


# MULTITHREAD WRAPPERS
# -----------------------------------------------------------------------------

def write_seqrecords(seqrecord_list, file_format, export_path,
                     export_name=None, concatenate=False, threads=1,
                     verbose=False):
    """Outputs files with a particuar format from a SeqRecord list.

    :param seq_record_list: List of populated SeqRecords.
    :type seq_record_list: list[SeqRecord]
    :param file_format: Biopython supported file type.
    :type file_format: str
    :param export_path: Path to a dir for file creation.
    :type export_path: Path
    :param concatenate: A boolean to toggle concatenation of SeqRecords.
    :type concaternate: bool
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    """
    record_dictionary = {}
    if concatenate:
        if export_name is None:
            record_dictionary.update({export_path.name: seqrecord_list})
        else:
            record_dictionary.update({export_name: seqrecord_list})
    else:
        for record in seqrecord_list:
            record_dictionary.update({record.name: record})

    if verbose:
        print(f"Writing selected data to files at '{export_path}'...")

    work_items = []
    for record_name, records in record_dictionary.items():
        file_name = f"{record_name}.{file_format}"
        file_path = export_path.joinpath(file_name)

        work_items.append((records, file_path, file_format))
    multithread.multithread(work_items, threads, write_seqrecord,
                            verbose=verbose)



from cmd import Cmd
from pathlib import Path
import shlex
import textwrap

from urllib3.exceptions import RequestError

from pdm_utils.functions import (basic, url_basic)


# SHELL_INTROS
# -----------------------------------------------------------------------------
GET_DB_CMD_INTRO = ("Loaded the pdm_utils get_db interactive environment:\n\n"
                    "Type select [package] to choose a database to install\n"
                    "Use cd [path/to/directory] or ls [path/to/directory] to "
                    "navigate remote contents\n"
                    "Type exit to leave the interactive environment\n\n"
                    "Type help or ? to list more commands.")


# GET_DB_SHELL
# -----------------------------------------------------------------------------
GET_DB_FILE_CMD_COMMANDS = ["desc", "select", "version"]
GET_DB_DIR_CMD_COMMANDS = ["cd"]
GET_DB_PATH_CMD_COMMANDS = (GET_DB_FILE_CMD_COMMANDS +
                            GET_DB_DIR_CMD_COMMANDS + ["ls"])
GET_DB_ARGLESS_COMMANDS = ["exit", "reload"]
GET_DB_SINGLE_ARG_COMMANDS = ["which", "load"]
GET_DB_CMD_COMMANDS = (GET_DB_PATH_CMD_COMMANDS + GET_DB_ARGLESS_COMMANDS +
                       GET_DB_SINGLE_ARG_COMMANDS)


# Inherits from Cmd library:
# Command line interactive faux shell that allows for database selection at
# some url
class GetDBCMD(Cmd):
    def __init__(self, url, name="root", preload=False, pool=None, **kwargs):
        super().__init__(**kwargs)

        if pool is None:
            pool = url_basic.create_pool(pipeline=True)

        self.pool = pool
        self.name = name
        self.preload = preload
        self.load_graph(url)

        self.selected = None
        self.tmp = None
        self.text_wrapper = textwrap.TextWrapper(width=80, initial_indent="\t")

        self.doc_header = ("Commands: (type help [command] for more info)")

    def set_prompt(self):
        prompt = "".join([self.name, "@", self.url])

        if self.dir_graph is not None:
            prompt = "".join([prompt, "::",
                              self.dir_graph.curr.get_rel_path()])

        self.prompt = "".join([prompt, "$ "])

    def load_graph(self, url):
        self.url = url

        self.dir_graph = None
        try:
            self.dir_graph = url_basic.UrlPathGraph(
                                            url, preload=True, pool=self.pool)
        except RequestError as e:
            print(e)
            return

        if self.preload:
            for node in self.dir_graph.nodes:
                if not node.isdir:
                    self.load_yaml_data(node)

    def load_yaml_data(self, node):
        for file_name in node.contents:
            if file_name.endswith(".yml"):
                file_url = "".join([node.get_abs_path(), "/", file_name])

                try:
                    response = url_basic.pool_request(file_url,
                                                      pool=self.dir_graph.pool)
                except RequestError as e:
                    print(e)

                node.data = url_basic.get_yaml_data(response)

    def preloop(self):
        self.set_prompt()

    def precmd(self, line):
        split_args_list = shlex.split(line)
        if split_args_list:
            command = split_args_list[0]

            if command in GET_DB_PATH_CMD_COMMANDS:
                self.tmp = self.parse_path_command(
                                                    split_args_list, command)

            if command in GET_DB_SINGLE_ARG_COMMANDS:
                self.tmp = self.parse_single_arg_command(
                                                    split_args_list, command)
        return line

    def postcmd(self, stop, line):
        self.set_prompt()
        self.tmp = None
        return stop

    # This functionality could be replaced/improved with python's argparse
    # library, however argparse exits on error.  This functionality is
    # toggleable with python3 v3.9, but this is the most recent python3
    # version as of 01/01/2021.  Could be updated as this version becomes more
    # commonly used.
    def parse_path_command(self, split_args_list, command):
        if self.dir_graph is None:
            print("".join(["Request failed.  "
                           "Please check the availability at '", self.url,
                           "'\nLoad a new url with the following:\n"
                           "\tload [url]"]))
            return None

        if len(split_args_list) == 1:
            path = ""
        else:
            path = split_args_list[1]

        if command in GET_DB_FILE_CMD_COMMANDS:
            node_type = "file"
        elif command in GET_DB_DIR_CMD_COMMANDS:
            node_type = "dir"
        else:
            node_type = None

        return self.get_path_node(path, command, node_type)

    def parse_single_arg_command(self, split_args_list, command):
        if len(split_args_list) == 1:
            arg = ""
        else:
            arg = split_args_list[1]

        return arg

    def get_path_node(self, path, command="", node_type=None):
        corrected_path = str(Path(path))
        path_node = self.dir_graph.traverse(corrected_path)

        if path_node is not None:
            if node_type == "file":
                if path_node.isdir:
                    print("".join([command, ": ", path, ": Not a package"]))
                    path_node = None
            elif node_type == "dir":
                if not path_node.isdir:
                    print("".join([command, ": ", path, ": Not a directory"]))
                    path_node = None
        else:
            print("".join([command, ": ", path,
                           ": Not a package or directory"]))

        return path_node

    def do_ls(self, arg):
        ("List directory contents\n"
         "\tUsage: ls [path/to/dir]")
        node = self.tmp
        if node is not None:
            if node.isdir:
                node.print_children()
            else:
                print(node.get_rel_path())

    def do_cd(self, arg):
        ("Change directory\n"
         "\tUsage: cd [path/to/dir]")
        node = self.tmp
        if node is not None:
            self.dir_graph.curr = node

    def do_desc(self, arg):
        ("Describe database package\n"
         "\tUsage: desc [path/to/package]")
        node = self.tmp
        if node is not None:
            if node.data is None:
                self.load_yaml_data(node)

            if node.data is not None:
                lines = []
                if node.data.get("name") is not None:
                    lines.append("Name:")
                    lines += self.text_wrapper.wrap(node.data["name"])

                if node.data.get("date") is not None:
                    lines.append("Date:")
                    lines += self.text_wrapper.wrap(str(node.data["date"]))

                if node.data.get("description") is not None:
                    lines.append("Description:")
                    lines += self.text_wrapper.wrap(node.data["description"])

                for line in lines:
                    print(line)

    def do_version(self, arg):
        ("Version database package\n"
         "\tUsage: version [path/to/package]")
        node = self.tmp
        if node is not None:
            if node.data is None:
                self.load_yaml_data(node)

            if node.data is not None:
                print(": ".join([
                        "Database version", str(node.data.get("version"))]))

    def do_select(self, arg):
        ("Select database package\n"
         "\tUsage: select [path/to/package]")
        node = self.tmp
        if node is not None:
            self.selected = node
            return True

    def do_which(self, arg):
        ("Locate a directory/package\n"
         "\tUsage: which [name]")
        arg = self.tmp
        for node in self.dir_graph.nodes:
            if node.name == arg:
                print(node.get_rel_path())

    def do_clear(self, arg):
        "Bring the command line to the top of screen"
        basic.clear_screen()

    def do_reload(self, arg):
        ("Reload root contents at the current url")
        self.load_graph(self.dir_graph.url)

    def do_load(self, arg):
        ("Reload root contents from a new url\n"
         "\tUsage: load [url]")
        self.load_graph(self.tmp)

    def do_exit(self, arg):
        ("Exits the interactive shell")
        return True



"""Object to provide a formatted filtering query
for retrieving data from a SQL database."""
from networkx import Graph
from sqlalchemy import Column
from sqlalchemy import and_
from sqlalchemy import or_
from sqlalchemy.engine.base import Engine
from sqlalchemy.orm.decl_api import DeclarativeMeta
from sqlalchemy.sql.elements import BooleanClauseList
from sqlalchemy.sql.elements import Label
from sqlalchemy.orm.session import Session

from pdm_utils.classes.alchemyhandler import AlchemyHandler
from pdm_utils.functions import basic
from pdm_utils.functions import cartography
from pdm_utils.functions import parsing
from pdm_utils.functions import querying as q

# -----------------------------------------------------------------------------
# GLOBAL VARIABLES

COLUMN_TYPES = [Column, Label]


class Filter:
    def __init__(self, alchemist=None, key=None):
        self._engine = None
        self._graph = None
        self._session = None
        self._mapper = None

        self._connected = False

        if isinstance(alchemist, AlchemyHandler):
            if not alchemist.connected:
                alchemist.connect(ask_database=True)

            self._engine = alchemist.engine
            self._graph = alchemist.graph
            self._session = alchemist.session
            self._mapper = alchemist.mapper

            self._connected = True

        if isinstance(key, Column):
            self._key = key
        else:
            self._key = None

        self._values = []
        self._values_valid = True

        self._filters = []
        self._updated = True
        self._or_index = -1

        self.verbose = False

# -----------------------------------------------------------------------------
# FILTER CONNECTION HANDLING

    def connect(self, alchemist=None):
        """Connect Filter object to a database with an AlchemyHandler.

        :param alchemist: An AlchemyHandler object.
        :type alchemist: AlchemyHandler
        """
        if alchemist is not None:
            self.link(alchemist)
            return

        if self._connected:
            return

        alchemist = AlchemyHandler()
        alchemist.connect(ask_database=True)

        self._engine = alchemist.engine
        self._graph = alchemist.graph
        self._session = alchemist.session
        self._mapper = alchemist.mapper

        self._connected = True

    def link(self, alchemist):
        """Connect Filter object to a database with an existing AlchemyHandler.

        :param alchemist: An AlchemyHandler object.
        :type alchemist: AlchemyHandler
        """
        if not isinstance(alchemist, AlchemyHandler):
            raise TypeError("AlchemyHandler to be linked is not "
                            "of type AlchemyHandler.")

        if not alchemist.connected:
            alchemist.connect(ask_database=True)

        self._engine = alchemist.engine
        self._graph = alchemist.graph
        self._session = alchemist.session
        self._mapper = alchemist.mapper

        self._connected = True

    def check(self):
        """Check Filter object contains valid essential objects.
        Filter object requires a SQLAlchemy Engine and Column as well as
        a NetworkX Graph.
        """
        if not self._connected:
            self.connect()
        else:
            if not isinstance(self._engine, Engine):
                raise AttributeError("Filter object is missing valid Engine.")
            if not isinstance(self._graph, Graph):
                raise AttributeError("Filter object is missing valid Graph.")
            if not isinstance(self._session, Session):
                raise AttributeError("Filter object is missing valid Session.")

        if not isinstance(self._key, Column):
            raise AttributeError("Filter object is missing valid column key.")

# -----------------------------------------------------------------------------
# FILTER CONNECTION HANDLING PROPERTIES

    @property
    def connected(self):
        connected = self._connected
        return connected

    @property
    def engine(self):
        engine = self._engine
        return engine

    @property
    def graph(self):
        graph = self._graph
        return graph

    @property
    def session(self):
        session = self._session
        return session

    @property
    def mapper(self):
        mapper = self._mapper
        return mapper

# -----------------------------------------------------------------------------
# FILTER VALUE HANDLING

    def refresh(self):
        """Re-queries for the Filter's values.
        """
        self.check()

        if self._values_valid:
            return

        values = self.build_values()
        self._values = values
        self._values_valid = True

    def update(self):
        """Queries using the Filter's key and its stored BinaryExpressions.
        """
        self.check()

        if not self._values_valid:
            self.refresh()

        if self._updated:
            return

        where_clauses = self.build_where_clauses()
        values = self.build_values(where=where_clauses)
        self._values = values

        self._updated = True
        self._values_valid = True

    def sort(self, raw_columns):
        """Re-queries for the Filter's values, applying a ORDER BY clause.

        :param raw_column: SQLAlchemy Column object(s) or object name(s).
        :type raw_columns: Column
        :type raw_columns: str
        :type raw_columns: list[Column]
        :type raw_columns: list[str]
        """
        self.check()

        columns = self.get_columns(raw_columns)

        query = q.build_select(self._graph, self._key, order_by=columns)

        values = q.first_column(self._engine, query, in_column=self._key,
                                values=self._values)
        self._values = values
        self._values_valid = True

# -----------------------------------------------------------------------------
# FILTER VALUE HANDLING PROPERTIES

    @property
    def values(self):
        if self._values is not None:
            values = self._values.copy()
            return values
        else:
            return []

    @values.setter
    def values(self, values):
        if not (isinstance(values, list) or values is None):
            raise TypeError("Filter object values must be of type list.")
        self._values = values
        self._values_valid = False
        self._updated = False

    @property
    def key(self):
        key = self._key
        return key

    @key.setter
    def key(self, key):
        if isinstance(key, Column):
            self._key = key
        elif isinstance(key, str):
            if self.graph is None:
                raise ValueError("String key input requires MySQL connection.")

            metadata = self.graph.graph["metadata"]

            try:
                self._key = q.get_column(self.graph.graph["metadata"], key)
            except:
                try:
                    table_obj = q.get_table(metadata, key)
                except:
                    raise ValueError("Inputted string key is neither a valid "
                                     "MySQL column or table.")

                self._key = list(table_obj.primary_key.columns)[0]

        else:
            raise TypeError("Filter key value is invalid."
                            "Filter key must be one of the following: \n"
                            "SQLAlchemy Column\n"
                            "MySQL column string\n"
                            "MySQL table string\n")

    @property
    def values_valid(self):
        values_valid = self._values_valid
        return values_valid

# FILTER CLAUSE HANDLING
# -----------------------------------------------------------------------------

    def reset(self):
        """Resets all filters, values, and Filter state conditions.
        """
        self.reset_filters()

        self._values = []
        self._values_valid = True

    def reset_filters(self):
        """Resets all filters and relevant Filter state condition.
        """
        self._filters = []
        self._updated = True
        self._or_index = -1

    def parenthesize(self):
        """Condense current filters into an isolated clause"""
        if self._or_index > 0:
            conditionals = self.build_where_clauses()

            self.reset_filters()
            self.new_or_()
            or_block = self._filters[self._or_index]

            or_block.update({"parenthetical": conditionals})

            self._updated = False

    def and_(self, filter):
        """Add an and conditional to the Filter object class.

        :param_filter: Formatted MySQL WHERE clause.
        :type_filter: str
        """
        if self._or_index < 0:
            self.new_or_()

        where_clause = q.build_where_clause(self.graph, filter)
        filter_key = parsing.create_filter_key(filter)

        or_block = self._filters[self._or_index]

        # Stores the BinaryExpression with a key of the Column/Operator pairing
        or_block.update({filter_key: where_clause})

        self._updated = False

    def new_or_(self):
        """Create a new conditional block to the Filter object class.
        """
        if self._or_index >= 0:
            if not self._filters[self._or_index]:
                return

        self._filters.append({})
        self._or_index += 1

    def remove(self, filter):
        """Remove an and filter from the current block of and conditionals.

        :param filter: Formatted MySQL WHERE clause.
        :type filter: str
        """
        filter_key = parsing.create_filter_key(filter)

        or_block = self._filters[self._or_index]

        # Indexes into the dictionary using the Column/Operator pairing.
        if filter_key in or_block.keys():
            or_block.pop(filter_key)
        else:
            raise IndexError(f"Filter {filter} is not within the current "
                             "block of and conditionals.")

        self._updated = False

    def add(self, filter_string):
        """Add a MySQL where filter(s) to the Filter object class.

        :param filter: Formatted MySQL WHERE clause.
        :type filter: str
        """
        filters = parsing.parse_cmd_string(filter_string)
        while(filters):
            and_filters = filters[0]
            for filter in and_filters:
                self.and_(filter)

            filters.pop(0)

            if filters:
                self.new_or_()

    def build_where_clauses(self):
        """Builds BinaryExpression objects from stored Filter object filters.

        :returns: A list of SQLAlchemy WHERE conditionals.
        :rtype: list
        """
        and_conditionals = []

        for or_block in self._filters:
            where_clauses = []
            for filter_key in or_block.keys():
                where_clauses.append(or_block[filter_key])

            if where_clauses:
                and_conditionals.append(and_(*where_clauses))

        if len(and_conditionals) > 1:
            conditionals = or_(*and_conditionals)
        elif len(and_conditionals) == 1:
            conditionals = and_conditionals
        else:
            conditionals = []

        return conditionals

# -----------------------------------------------------------------------------
# FILTER CLAUSE HANDLING PROPERTIES

    @property
    def filters(self):
        filters = self.copy_filters()
        return filters

    @property
    def updated(self):
        updated = self._updated
        return updated

    @property
    def or_index(self):
        or_index = self._or_index
        return or_index

    @or_index.setter
    def or_index(self, index):
        or_count = len(self._filters)
        if index < 0 or index > (or_count - 1):
            raise IndexError("Or_index is out of bounds for "
                             "the current or_ blocks.\n"
                             f"Current or_ blocks stored are {or_count}.")

        self._or_index = index

# -----------------------------------------------------------------------------
# FILTER QUERYING
    def build_values(self, where=None, column=None, raw_bytes=False,
                     limit=8000):
        """Queries for values from stored WHERE clauses and Filter key.

        :param where: MySQL WHERE clause_related SQLAlchemy object(s).
        :type where: BinaryExpression
        :type where: list
        :param order_by: MySQL ORDER BY clause-related SQLAlchemy object(s).
        :type order_by: Column
        :type order_by: list
        :param column: SQLAlchemy Column object or object name.
        :type column: Column
        :type column: str
        :param limit: SQLAlchemy IN clause query length limiter.
        :type limit: int
        :returns: Distinct values fetched from given and innate constraints.
        :rtype: list
        """
        self.check()

        if column is None:
            column_obj = self._key
        else:
            column_obj = self.get_column(column)

        if where is not None:
            if isinstance(where, list) or isinstance(where, BooleanClauseList):
                base_clauses = where
            else:
                base_clauses = [where]
        else:
            base_clauses = []

        query = q.build_distinct(self.graph, column_obj, where=base_clauses,
                                 add_in=self._key)

        values = q.first_column(self.engine, query, in_column=self._key,
                                values=self._values, limit=limit)

        if not raw_bytes:
            if column_obj.type.python_type is bytes:
                values = basic.convert_to_decoded(values)

        return values

    def select(self, raw_columns, return_dict=True):
        """Queries for data conditioned on the values in the Filter object.

        :param columns: SQLAlchemy Column object(s)
        :type columns: Column
        :type columns: str
        :type columns: list[Column]
        :type columns: list[str]
        :param return_dict: Toggle whether to return data as a dictionary.
        :type return_dict: Boolean
        :returns: SELECT data conditioned on the values in the Filter object.
        :rtype: dict
        :rtype: list[RowProxy]
        """
        self.check()

        columns = self.get_columns(raw_columns)

        query = q.build_select(self._graph, columns, add_in=self._key)
        results = q.execute(self._engine, query, in_column=self._key,
                            values=self._values, return_dict=return_dict)

        return results

    def query(self, table_map):
        """Queries for ORM object instances conditioned on Filter values.

        :param table_map: SQLAlchemy ORM map object.
        :type table_map: str
        :type table_map: DeclarativeMeta
        :returns: List of mapped object instances.
        :rtype: list
        """
        self.check()

        if not self._values:
            return []

        if isinstance(table_map, str):
            table_map = cartography.get_map(self._mapper, table_map)
        elif isinstance(table_map, DeclarativeMeta):
            pass
        else:
            raise TypeError("Table map object must be either "
                            "type str or DeclarativeMeta.")

        in_clause = self._key.in_(self._values)

        instances = q.query(self._session, self._graph, table_map,
                            where=in_clause)
        return instances

    def transpose(self, raw_column, return_dict=False, set_values=False,
                  raw_bytes=False, filter=False):
        """Queries for distinct values from stored values and a MySQL Column.

        :param raw_column: SQLAlchemy Column object or object name.
        :type raw_column: Column
        :type raw_column: str
        :param return_dict: Toggle whether to return data as a dictionary.
        :type return_dict: Boolean
        :param set_values: Toggle whether to replace Filter key and values.
        :type set_values: Boolean
        :returns: Distinct values fetched from given and innate constraints.
        :rtype: list
        :rtype: dict
        """
        self.check()

        if not self._values:
            if return_dict:
                return {}
            else:
                return []

        column = self.get_column(raw_column)
        name = column.name

        where_clauses = None
        if filter:
            where_clauses = self.build_where_clauses()

        values = self.build_values(column=column, where=where_clauses,
                                   raw_bytes=raw_bytes)

        if set_values:
            self._key = column
            self._values = values

        if return_dict:
            values = {name: values}

        return values

    def mass_transpose(self, raw_columns, raw_bytes=False, filter=False):
        """Queries for sets of distinct values, using self.transpose()

        :param columns: SQLAlchemy Column object(s)
        :type columns: Column
        :type columns: list
        :returns: Distinct values fetched from given and innate restraints.
        :rtype: dict
        """
        self.check()

        if not self._values:
            return {}

        if not isinstance(raw_columns, list):
            raw_columns = [raw_columns]

        values = {}
        for column in raw_columns:
            column_values = self.transpose(column, return_dict=True,
                                           raw_bytes=raw_bytes,
                                           filter=filter)
            values.update(column_values)

        return values

    def group(self, raw_column, raw_bytes=False, filter=False):
        """Queries and separates Filter object's values based on a Column.

        :param raw_column: SQLAlchemy Column object or object name.
        :type raw_column: Column
        :type raw_column: str
        """
        self.check()

        column = self.get_column(raw_column)

        groups = self.transpose(column)

        where_clauses = []
        if filter:
            where_clauses = self.build_where_clauses()

        group_results = {}
        for group in groups:
            if column.type.python_type == bytes:
                group_clauses = where_clauses + \
                                            [(column == group.encode("utf_8"))]
            else:
                group_clauses = where_clauses + [(column == group)]

            values = self.build_values(where=group_clauses,
                                       raw_bytes=raw_bytes)
            group_results.update({group: values})

        return group_results

    def retrieve(self, raw_columns, raw_bytes=False, filter=False):
        """Queries for distinct data for each value in the Filter object.

        :param columns: SQLAlchemy Column object(s)
        :type columns: Column
        :type columns: str
        :type columns: list[Column]
        :type columns: list[str]
        :returns: Distinct values for each Filter value.
        :rtype: dict{dict}
        """
        self.check()

        if not self._values:
            return {}

        columns = self.get_columns(raw_columns)

        where_clauses = []
        if filter:
            where_clauses = self.build_where_clauses()

        values = {}
        for value in self._values:
            compare_value = value
            if self._key.type.python_type == bytes and value is not None:
                compare_value = value.encode("utf-8")

            value_clauses = where_clauses + [(self._key == compare_value)]
            values.update({value: {}})

            # For each column for each value, add the respective data to a dict
            for i in range(len(columns)):
                query = q.build_distinct(self._graph, columns[i],
                                         where=value_clauses)
                value_data = q.first_column(self._engine, query)

                if not raw_bytes:
                    if columns[i].type.python_type == bytes:
                        value_data = basic.convert_to_decoded(value_data)

                values[value].update({columns[i].name: value_data})

        return values

    def get_column(self, raw_column):
        """Converts a column input, string or Column, to a Column.

        :param raw_column: SQLAlchemy Column object or object name.
        :type raw_column: Column
        :type raw_column: str
        """
        self.check()

        if isinstance(raw_column, str):
            column = q.get_column(self.graph.graph["metadata"], raw_column)
        elif type(raw_column) in COLUMN_TYPES:
            column = raw_column
        else:
            raise TypeError(
                        "Column must be either a string or a Column object")

        return column

    def get_columns(self, raw_columns):
        """Converts a column input list, string or Column, to a list of Columns.

        :param raw_column: SQLAlchemy Column object or object name.
        :type raw_column: list[Column]
        :type raw_column: list[str]
        :returns: Returns SQLAlchemy Columns
        :rtype: list[Column]
        """
        if not isinstance(raw_columns, list):
            raw_columns = [raw_columns]

        columns = []
        for raw_column in raw_columns:
            columns.append(self.get_column(raw_column))

        return columns

# -----------------------------------------------------------------------------
# FILTER QUALITY-OF-LIFE

    def hits(self):
        """Gets the number of a Filter object's values.
        """
        if self.values is None:
            return 0

        return len(self._values)

    def copy(self):
        """Returns a copy of a Filter object.
        """
        copy = Filter()
        copy._connected = self._connected
        copy._updated = self.updated
        copy._values_valid = self.values_valid
        copy._filters = self.copy_filters()
        copy._engine = self._engine
        copy._graph = self._graph
        copy._session = self._session
        copy._key = self.key
        copy._values = self.values

        return copy

    def copy_filters(self):
        """Returns a copy of a Filter object's filter dictionary.
        """
        filters_copy = []
        # Copies of each list are required so that multiple filter copies
        # are not adding and removing from the same space in memory.
        for or_block in self._filters:
            new_block = {}

            for clause in or_block.keys():
                new_block.update({clause: or_block[clause]})

            filters_copy.append(new_block)

        return filters_copy

    def print_results(self):
        """Prints the Filter object's values in a formatted way.
        """
        if not self._values:
            return

        if not self._values_valid:
            self.refresh()

        print(" " + "_"*57 + " ")
        print("| %-20s" %
              (f"{self.key.name} Results:")
              + " "*36 + "|")
        print("|" + "-"*57 + "|")
        for row in range(0, len(self._values), 3):
            result_row = self._values[row:row+3]
            if len(result_row) == 3:
                print("| %-20s %-20s %-13s |" %
                      (result_row[0], result_row[1], result_row[2]))
            elif len(result_row) == 2:
                print("| %-20s %-20s %-13s" %
                      (result_row[0], result_row[1], "")
                      + " " + "|")

            elif len(result_row) == 1:
                print("| %-20s %-20s %-13s" %
                      (result_row[0], "", "")
                      + " " + "|")
        print("|" + "_"*57 + "|")



"""Represents a structure to directly compare data between two or more genomes."""

from pdm_utils.classes import evaluation
from pdm_utils.classes import ticket
from pdm_utils.functions import basic

class Bundle:

    # Initialize all attributes:
    def __init__(self):

        # Initialize all non-calculated attributes:
        self.id = ""
        self.ticket = None
        self.genome_dict = {}
        self.genome_pair_dict = {}
        self.email = None # Email object for automating responses


        # Initialize calculated attributes
        self.evaluations = []
        self.sql_statements = [] # All SQL data needed to implement ticket.

        self._errors = 0





    # TODO should this be changed to generate the GenomePair object directly?
    def set_genome_pair(self, genome_pair, key1, key2):
        """Pair two genomes and add to the paired genome dictionary.

        :param genome_pair:
            An empty GenomePair object to stored paried genomes.
        :type genome_pair: GenomePair
        :param key1:
            A valid key in the Bundle object's 'genome_dict'
            that indicates the first genome to be paired.
        :type key1: str
        :param key2:
            A valid key in the Bundle object's 'genome_dict'
            that indicates the second genome to be paired.
        :type key2: str
        """

        try:
            genome_pair.genome1 = self.genome_dict[key1]
            genome_pair.genome2 = self.genome_dict[key2]
            paired_key = key1 + "_" + key2
            self.genome_pair_dict[paired_key] = genome_pair
        except:
            pass



    # Evaluations.
    def set_eval(self, eval_id, definition, result, status):
        """Constructs and adds an Evaluation object to the evaluations list.

        :param eval_id: Unique identifier for the evaluation.
        :type eval_id: str
        :param definition: Description of the evaluation.
        :type definition: str
        :param result: Description of the outcome of the evaluation.
        :type result: str
        :param status: Outcome of the evaluation.
        :type status: str
        """
        evl = evaluation.Evaluation(eval_id, definition, result, status)
        self.evaluations.append(evl)

    def check_ticket(self, eval_id=None, success="correct", fail="error",
                     eval_def=None):
        """Check for whether a Ticket object is present.

        :param eval_id: same as for check_genome_dict().
        :param success: same as for check_genome_dict().
        :param fail: same as for check_genome_dict().
        :param eval_def: same as for check_genome_dict().
        """
        if self.ticket is not None:
            result = ("A ticket is present. "
                      f"ID: {self.ticket.id}. "
                      f"Type: {self.ticket.type}. "
                      f"PhageID: {self.ticket.phage_id}.")
            status = success
        else:
            result = "A ticket is not present."
            status = fail
        definition = "Check if a ticket is present."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_genome_dict(self, key, expect=True, eval_id=None,
                          success="correct", fail="error", eval_def=None):
        """Check if a genome is present in the genome dictionary.

        :param key:
            The value to be evaluated if it is a valid key
            in the genome dictionary.
        :type key: str
        :param expect:
            Indicates whether the key is expected
            to be a valid key in the genome dictionary.
        :type expect: bool
        :param eval_id: Unique identifier for the evaluation.
        :type eval_id: str
        :param success: Default status if the outcome is a success.
        :type success: str
        :param fail: Default status if the outcome is not a success.
        :type fail: str
        :param eval_def: Description of the evaluation.
        :type eval_def: str
        """
        if key in self.genome_dict.keys():
            result = f"The '{key}' genome is present."
            if expect:
                status = success
            else:
                status = fail
        else:
            result = f"The '{key}' genome is not present."
            if not expect:
                status = success
            else:
                status = fail

        definition = "Check if a genome is present."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_genome_pair_dict(self, key, expect=True, eval_id=None,
                               success="correct", fail="error", eval_def=None):
        """Check if a genome_pair is present in the genome_pair dictionary.

        :param key: same as for check_genome_dict().
        :param expect: same as for check_genome_dict().
        :param eval_id: same as for check_genome_dict().
        :param success: same as for check_genome_dict().
        :param fail: same as for check_genome_dict().
        :param eval_def: same as for check_genome_dict().
        """

        if key in self.genome_pair_dict.keys():
            result = f"The '{key}' genome_pair is present."
            if expect:
                status = success
            else:
                status = fail
        else:
            result = f"The '{key}' genome_pair is not present."
            if not expect:
                status = success
            else:
                status = fail

        definition = "Check if a genome_pair is present."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, result, status)

    def check_statements(self, execute_result, execute_msg, eval_id=None,
                         success="correct", fail="error", eval_def=None):
        """Check if MySQL statements were successfully executed.

        :param execute_result:
            Indication if MySQL statements were successfully execute.
        :type execute_result: int
        :param execute_msg:
            Description of MySQL statement execution result.
        :type execute_msg: str
        :param eval_id: same as for check_genome_dict().
        :param success: same as for check_genome_dict().
        :param fail: same as for check_genome_dict().
        :param eval_def: same as for check_genome_dict().
        """
        if execute_result == 0:
            status = success
        else:
            status = fail
        definition = "Check if MySQL statements were successfully executed."
        definition = basic.join_strings([definition, eval_def])
        self.set_eval(eval_id, definition, execute_msg, status)

    def check_for_errors(self):
        """Check evaluation lists of all objects contained in the Bundle
        and determine how many errors there are."""
        # TODO eventually, cds_pair objects can also be evaluated here, once
        # they are integrated more fully into pdm_utils.

        self._errors = 0
        for evl in self.evaluations:
            if evl.status == "error":
                self._errors += 1

        if self.ticket is not None:
            for evl in self.ticket.evaluations:
                if evl.status == "error":
                    self._errors += 1

        for key in self.genome_dict.keys():
            gnm = self.genome_dict[key]
            for evl in gnm.evaluations:
                if evl.status == "error":
                    self._errors += 1

            for src_ftr in gnm.source_features:
                for evl in src_ftr.evaluations:
                    if evl.status == "error":
                        self._errors += 1

            for cds_ftr in gnm.cds_features:
                for evl in cds_ftr.evaluations:
                    if evl.status == "error":
                        self._errors += 1

            for trna_ftr in gnm.trna_features:
                for evl in trna_ftr.evaluations:
                    if evl.status == "error":
                        self._errors += 1

            for tmrna_ftr in gnm.tmrna_features:
                for evl in tmrna_ftr.evaluations:
                    if evl.status == "error":
                        self._errors += 1

        for key in self.genome_pair_dict.keys():
            genome_pair = self.genome_pair_dict[key]
            for evl in genome_pair.evaluations:
                if evl.status == "error":
                    self._errors += 1


    def get_evaluations(self):
        """Get all evaluations for all objects stored in the Bundle.

        :returns: Dictionary of evaluation lists for each feature.
        :rtype: dict
        """
        eval_dict = {}
        if len(self.evaluations) > 0:
            eval_dict["bundle"] = self.evaluations
        if isinstance(self.ticket, ticket.ImportTicket):
            if len(self.ticket.evaluations) > 0:
                eval_dict["ticket"] = self.ticket.evaluations
        for key in self.genome_dict.keys():
            gnm = self.genome_dict[key]
            genome_key = "genome_" + key
            if len(gnm.evaluations) > 0:
                eval_dict[genome_key] = gnm.evaluations

            for src_ftr in gnm.source_features:
                src_key = "src_" + src_ftr.id
                if len(src_ftr.evaluations) > 0:
                    eval_dict[src_key] = src_ftr.evaluations

            for cds_ftr in gnm.cds_features:
                cds_key = "cds_" + cds_ftr.id
                if len(cds_ftr.evaluations) > 0:
                    eval_dict[cds_key] = cds_ftr.evaluations

            for trna_ftr in gnm.trna_features:
                trna_key = "trna_" + trna_ftr.id
                if len(trna_ftr.evaluations) > 0:
                    eval_dict[trna_key] = trna_ftr.evaluations

            for tmrna_ftr in gnm.tmrna_features:
                tmrna_key = "tmrna_" + tmrna_ftr.id
                if len(tmrna_ftr.evaluations) > 0:
                    eval_dict[tmrna_key] = tmrna_ftr.evaluations

        for key in self.genome_pair_dict.keys():
            genome_pair = self.genome_pair_dict[key]
            genome_pair_key = "genome_pair_" + key
            if len(genome_pair.evaluations) > 0:
                eval_dict[genome_pair_key] = genome_pair.evaluations
        return eval_dict



"""Pipeline to update specific fields in a MySQL database."""
import argparse
import csv
import pathlib

from sqlalchemy import update

from pdm_utils.classes.alchemyhandler import AlchemyHandler
from pdm_utils.functions import basic
from pdm_utils.functions import configfile
from pdm_utils.functions import mysqldb
from pdm_utils.functions import mysqldb_basic
from pdm_utils.functions import querying


# TODO unittest.
def main(unparsed_args):
    """Runs the complete update pipeline."""
    args = parse_args(unparsed_args[2:])

    # Verify database connection and schema compatibility.
    print("Connecting to the MySQL database...")

    # Create config object with data obtained from file and/or defaults.
    config = configfile.build_complete_config(args.config_file)
    mysql_creds = config["mysql"]
    alchemist = AlchemyHandler(database=args.database,
                               username=mysql_creds["user"],
                               password=mysql_creds["password"])
    alchemist.connect(pipeline=True)
    engine = alchemist.engine
    mysqldb.check_schema_compatibility(engine, "the update pipeline")

    if args.version is True:
        mysqldb.change_version(engine)
        print("Database version updated.")

    if args.ticket_table is not None:
        update_table_path = basic.set_path(args.ticket_table,
                                kind="file", expect=True)

        # Iterate through the tickets and process them sequentially.
        list_of_update_tickets = []
        with update_table_path.open(mode='r') as f:
            file_reader = csv.DictReader(f)
            for dict in file_reader:
                list_of_update_tickets.append(dict)

        # Variables to be used for end summary
        processed = 0
        succeeded = 0
        failed = 0

        for dict in list_of_update_tickets:
            status = update_field(alchemist, dict)

            if status == 1:
                processed += 1
                succeeded += 1
            else:
                processed += 1
                failed += 1

        engine.dispose()
        print("\nDone iterating through tickets.")
        if succeeded > 0:
            print(f"{succeeded} / {processed} tickets successfully handled.")
        if failed > 0:
            print(f"{failed} / {processed} tickets failed to be handled.")


def update_field(alchemist, update_ticket):
    """Attempts to update a field using information from an update_ticket.

    :param alchemist: A connected and fully build AlchemyHandler object.
    :type alchemist: AlchemyHandler
    :param update_ticket: Dictionary with instructions to update a field.
    :type update_ticket: dict
    """
    try:
        table_map = alchemist.mapper.classes[update_ticket["table"]]
    except:
        print(f"\nInvalid table '{update_ticket['table']}'")
        return 0

    table_obj = table_map.__table__

    try:
        field = table_obj.c[update_ticket["field"]]
    except:
        print(f"\nInvalid replacement field '{update_ticket['field']}' "
              f"for table '{table_obj.name}'")
        return 0

    try:
        key_field = table_obj.c[update_ticket["key_name"]]
    except:
        print(f"\nInvalid selection key '{update_ticket['key_name']}' "
              f"for table '{table_obj.name}'")
        return 0

    primary_keys = list(table_obj.primary_key.columns)
    if key_field not in primary_keys:
        print(f"\nInvalid selection key '{update_ticket['key_name']}' "
              f"for table '{table_obj.name}'")
        return 0

    key_value_clause = (key_field == update_ticket["key_value"])

    key_value_query = querying.build_count(alchemist.graph, key_field, where=key_value_clause)
    key_value_count = mysqldb_basic.scalar(alchemist.engine, key_value_query)

    if key_value_count != 1:
        print(f"\nInvalid selection value '{update_ticket['key_value']}' "
              f"for key '{key_field.name}' in table '{table_obj.name}'")
        return 0

    if update_ticket["value"] == "NULL":
        update_ticket["value"] = None

    statement = update(table_obj).where(key_value_clause).values(
                                    {field.name : update_ticket["value"]})
    alchemist.engine.execute(statement)
    return 1


# TODO unittest.
def parse_args(unparsed_args_list):
    """Verify the correct arguments are selected for getting updates."""
    update_help = "Pipeline to update specific fields in a MySQL database."
    database_help = "Name of the MySQL database."
    ticket_table_help = """
        Path to the CSV-formatted table containing
        instructions to process each update.
        Structure of update ticket table:
            1. Table name
            2. Column name to be updated
            3. New value that will be added
            4. Conditional column name
            5. Conditional column value
        """
    version_help = "Increment database version by 1."
    config_file_help = "Path to the file containing user-specific login data."

    parser = argparse.ArgumentParser(description=update_help)
    parser.add_argument("database", type=str, help=database_help)
    parser.add_argument("-f", "--ticket_table", type=pathlib.Path,
                        help=ticket_table_help)
    parser.add_argument("-v", "--version", action="store_true",
                        default=False, help=version_help)
    parser.add_argument("-c", "--config_file", type=pathlib.Path,
                        help=config_file_help, default=None)
    args = parser.parse_args(unparsed_args_list)
    return args
